#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
import math

import numexpr as ne
import numpy as np
import pytest

import blosc2

NITEMS_SMALL = 1000
NITEMS = 10_000


@pytest.fixture(params=[np.float32, np.float64])
def dtype_fixture(request):
    return request.param


@pytest.fixture(params=[(NITEMS_SMALL,), (NITEMS,), (NITEMS // 100, 100)])
def shape_fixture(request):
    return request.param


@pytest.fixture
def array_fixture(dtype_fixture, shape_fixture):
    nelems = np.prod(shape_fixture)
    na1 = np.linspace(0, 10, nelems, dtype=dtype_fixture).reshape(shape_fixture)
    # For full generality, use different chunks and blocks
    # chunks = [c // 17 for c in na1.shape]
    # blocks = [c // 19 for c in na1.shape]
    # chunks1 = [c // 23 for c in na1.shape]
    # blocks1 = [c // 29 for c in na1.shape]
    chunks = [c // 4 for c in na1.shape]
    blocks = [c // 8 for c in na1.shape]
    chunks1 = [c // 10 for c in na1.shape]
    blocks1 = [c // 30 for c in na1.shape]
    a1 = blosc2.asarray(na1, chunks=chunks, blocks=blocks)
    na2 = np.copy(na1)
    a2 = blosc2.asarray(na2, chunks=chunks, blocks=blocks)
    na3 = np.copy(na1)
    # Let other operands have chunks1 and blocks1
    a3 = blosc2.asarray(na3, chunks=chunks1, blocks=blocks1)
    na4 = np.copy(na1)
    a4 = blosc2.asarray(na4, chunks=chunks1, blocks=blocks1)
    return a1, a2, a3, a4, na1, na2, na3, na4


# @pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all"])
@pytest.mark.parametrize("reduce_op", ["sum"])
def test_reduce_bool(array_fixture, reduce_op):
    a1, a2, a3, a4, na1, na2, na3, na4 = array_fixture
    expr = a1 + a2 > a3 * a4
    nres = ne.evaluate("na1 + na2 > na3 * na4")
    # res = getattr(expr, reduce_op)()
    res = expr.sum()
    # print("res:", res)
    nres = getattr(nres, reduce_op)()
    tol = 1e-15 if a1.dtype == "float64" else 1e-6
    np.testing.assert_allclose(res, nres, atol=tol, rtol=tol)


@pytest.mark.parametrize("reduce_op", ["sum", "prod", "mean", "std", "var", "min", "max", "any", "all"])
@pytest.mark.parametrize("axis", [1, (0, 1), None])
@pytest.mark.parametrize("keepdims", [True, False])
@pytest.mark.parametrize("dtype_out", [np.int16, np.float64])
@pytest.mark.parametrize(
    "kwargs",
    [{}, {"cparams": blosc2.CParams(clevel=1, filters=[blosc2.Filter.BITSHUFFLE], filters_meta=[0])}],
)
@pytest.mark.heavy
def test_reduce_params(array_fixture, axis, keepdims, dtype_out, reduce_op, kwargs):
    a1, a2, a3, a4, na1, na2, na3, na4 = array_fixture
    if axis is not None and np.isscalar(axis) and len(a1.shape) >= axis:
        return
    if isinstance(axis, tuple) and len(a1.shape) < len(axis):
        return
    if reduce_op == "prod":
        # To avoid overflow, create a1 and a2 with small values
        na1 = np.linspace(0, 0.1, np.prod(a1.shape), dtype=np.float32).reshape(a1.shape)
        a1 = blosc2.asarray(na1)
        na2 = np.linspace(0, 0.5, np.prod(a1.shape), dtype=np.float32).reshape(a1.shape)
        a2 = blosc2.asarray(na2)
        expr = a1 + a2 - 0.2
        nres = eval("na1 + na2 - .2")
    else:
        expr = a1 + a2 - a3 * a4
        nres = eval("na1 + na2 - na3 * na4")
    if reduce_op in ("sum", "prod", "mean", "std"):
        if reduce_op in ("mean", "std") and dtype_out == np.int16:
            # mean and std need float dtype as output
            dtype_out = np.float64
        res = getattr(expr, reduce_op)(axis=axis, keepdims=keepdims, dtype=dtype_out, **kwargs)
        nres = getattr(nres, reduce_op)(axis=axis, keepdims=keepdims, dtype=dtype_out)
    else:
        res = getattr(expr, reduce_op)(axis=axis, keepdims=keepdims, **kwargs)
        nres = getattr(nres, reduce_op)(axis=axis, keepdims=keepdims)
    tol = 1e-15 if a1.dtype == "float64" else 1e-6
    if kwargs != {}:
        if not np.isscalar(res):
            assert isinstance(res, blosc2.NDArray)
        np.testing.assert_allclose(res[()], nres, atol=tol, rtol=tol)
    else:
        np.testing.assert_allclose(res, nres, atol=tol, rtol=tol)


# TODO: "prod" is not supported here because it overflows with current values
@pytest.mark.parametrize("reduce_op", ["sum", "min", "max", "mean", "std", "var", "any", "all"])
@pytest.mark.parametrize("axis", [0, 1, None])
def test_reduce_expr_arr(array_fixture, axis, reduce_op):
    a1, a2, a3, a4, na1, na2, na3, na4 = array_fixture
    if axis is not None and len(a1.shape) >= axis:
        return
    expr = a1 + a2 - a3 * a4
    nres = eval("na1 + na2 - na3 * na4")
    res = getattr(expr, reduce_op)(axis=axis) + getattr(a1, reduce_op)(axis=axis)
    nres = getattr(nres, reduce_op)(axis=axis) + getattr(na1, reduce_op)(axis=axis)
    tol = 1e-15 if a1.dtype == "float64" else 1e-6
    np.testing.assert_allclose(res, nres, atol=tol, rtol=tol)


# Test broadcasting
@pytest.mark.parametrize("reduce_op", ["sum", "mean", "std", "var", "min", "max", "any", "all"])
@pytest.mark.parametrize("axis", [0, (0, 1), None])
@pytest.mark.parametrize("keepdims", [True, False])
@pytest.mark.parametrize(
    "shapes",
    [
        ((5, 5, 5), (5, 5), (5,)),
        ((10, 10, 10), (10, 10), (10,)),
        ((100, 100, 100), (100, 100), (100,)),
    ],
)
def test_broadcast_params(axis, keepdims, reduce_op, shapes):
    na1 = np.linspace(0, 1, np.prod(shapes[0])).reshape(shapes[0])
    na2 = np.linspace(1, 2, np.prod(shapes[1])).reshape(shapes[1])
    na3 = np.linspace(2, 3, np.prod(shapes[2])).reshape(shapes[2])
    a1 = blosc2.asarray(na1)
    a2 = blosc2.asarray(na2)
    a3 = blosc2.asarray(na3)

    expr1 = a1 + a2 - a3
    assert expr1.shape == shapes[0]
    expr2 = a1 * a2 + 1
    assert expr2.shape == shapes[0]
    res = expr1 - getattr(expr2, reduce_op)(axis=axis, keepdims=keepdims)
    assert res.shape == shapes[0]
    # print(f"res: {res.shape} expr1: {expr1.shape} expr2: {expr2.shape}")
    nres = eval(f"na1 + na2 - na3 - (na1 * na2 + 1).{reduce_op}(axis={axis}, keepdims={keepdims})")

    tol = 1e-14 if a1.dtype == "float64" else 1e-5
    np.testing.assert_allclose(res[:], nres, atol=tol, rtol=tol)


# Test reductions with item parameter
@pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all", "mean", "std", "var"])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("stripes", ["rows", "columns"])
@pytest.mark.parametrize("stripe_len", [2, 10, 15, 100])
@pytest.mark.parametrize("shape", [(10, 30), (30, 10), (50, 50)])
@pytest.mark.parametrize("chunks", [None, (10, 15), (20, 30)])
@pytest.mark.heavy
def test_reduce_item(reduce_op, dtype, stripes, stripe_len, shape, chunks):
    na = np.linspace(0, 1, num=np.prod(shape), dtype=dtype).reshape(shape)
    a = blosc2.asarray(na, chunks=chunks)
    tol = 1e-6 if dtype == np.float32 else 1e-15
    for i in range(0, a.shape[0], stripe_len):
        if stripes == "rows":
            _slice = (slice(i, i + stripe_len), slice(None))
        else:
            _slice = (slice(None), slice(i, i + stripe_len))
        slice_ = na[_slice]
        if slice_.size == 0 and reduce_op not in ("sum", "prod"):
            # For mean, std, and var, numpy just raises a warning, so don't check
            if reduce_op in ("min", "max"):
                # Check that a ValueError is raised when the slice is empty
                with pytest.raises(ValueError):
                    getattr(a, reduce_op)(item=_slice)
                with pytest.raises(ValueError):
                    getattr(na[_slice], reduce_op)()
        else:
            res = getattr(a, reduce_op)(item=_slice)
            nres = getattr(na[_slice], reduce_op)()
            np.testing.assert_allclose(res, nres, atol=tol, rtol=tol)


# Test fast path for reductions
@pytest.mark.parametrize(
    ("chunks", "blocks"),
    [
        ((20, 50, 100), (10, 50, 100)),
        ((10, 25, 70), (10, 25, 50)),
        ((10, 50, 100), (6, 25, 75)),
        ((15, 30, 75), (7, 20, 50)),
        ((20, 50, 100), (10, 50, 60)),
    ],
)
@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("fill_value", [0, 1, 0.32])
@pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all", "mean", "std", "var"])
@pytest.mark.parametrize("axis", [0, 1, None])
def test_fast_path(chunks, blocks, disk, fill_value, reduce_op, axis):
    shape = (20, 50, 100)
    urlpath = "a1.b2nd" if disk else None
    if fill_value != 0:
        a = blosc2.full(shape, fill_value, chunks=chunks, blocks=blocks, urlpath=urlpath, mode="w")
    else:
        a = blosc2.zeros(shape, dtype=np.float64, chunks=chunks, blocks=blocks, urlpath=urlpath, mode="w")
    if disk:
        a = blosc2.open(urlpath)
    na = a[:]

    res = getattr(a, reduce_op)(axis=axis)
    nres = getattr(na[:], reduce_op)(axis=axis)

    assert np.allclose(res, nres)


@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("fill_value", [0, 1, 0.32])
@pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all", "mean", "std", "var"])
@pytest.mark.parametrize("axis", [0, (0, 1), None])
def test_save_version1(disk, fill_value, reduce_op, axis):
    shape = (20, 50, 100)
    urlpath = "a1.b2nd" if disk else None
    if fill_value != 0:
        a = blosc2.full(shape, fill_value, urlpath=urlpath, mode="w")
        b = blosc2.full(shape, fill_value - 0.1, urlpath="b.b2nd", mode="w")
    else:
        a = blosc2.zeros(shape, dtype=np.float64, urlpath=urlpath, mode="w")
        b = blosc2.zeros(shape, dtype=np.float64, urlpath="b.b2nd", mode="w") - 0.1
    if disk:
        a = blosc2.open(urlpath)
        b = blosc2.open("b.b2nd")
    na = a[:]
    nb = b[:]

    # A reduction in the back
    expr = f"a + {reduce_op}(b, axis={axis}) + 1"
    lexpr = blosc2.lazyexpr(expr)
    assert lexpr.shape == a.shape
    if disk:
        lexpr.save("out.b2nd")
        lexpr = blosc2.open("out.b2nd")
    res = lexpr.compute()
    nres = na + getattr(nb[()], reduce_op)(axis=axis) + 1
    assert np.allclose(res[()], nres)

    if disk:
        blosc2.remove_urlpath("a1.b2nd")
        blosc2.remove_urlpath("b.b2nd")
        blosc2.remove_urlpath("out.b2nd")


@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("fill_value", [0, 1, 0.32])
@pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all", "mean", "std", "var"])
@pytest.mark.parametrize("axis", [0, (0, 1), None])
def test_save_version2(disk, fill_value, reduce_op, axis):
    shape = (20, 50, 100)
    urlpath = "a1.b2nd" if disk else None
    if fill_value != 0:
        a = blosc2.full(shape, fill_value, urlpath=urlpath, mode="w")
        b = blosc2.full(shape, fill_value - 0.1, urlpath="b.b2nd", mode="w")
    else:
        a = blosc2.zeros(shape, dtype=np.float64, urlpath=urlpath, mode="w")
        b = blosc2.zeros(shape, dtype=np.float64, urlpath="b.b2nd", mode="w") - 0.1
    if disk:
        a = blosc2.open(urlpath)
        b = blosc2.open("b.b2nd")
    na = a[:]
    nb = b[:]

    # A reduction in front
    expr = f"a.{reduce_op}(axis={axis}) + b"
    lexpr = blosc2.lazyexpr(expr, operands={"a": a, "b": b})
    if disk:
        lexpr.save("out.b2nd")
        lexpr = blosc2.open("out.b2nd")
    res = lexpr.compute()
    nres = getattr(na[()], reduce_op)(axis=axis) + nb
    assert np.allclose(res[()], nres)

    if disk:
        blosc2.remove_urlpath("a1.b2nd")
        blosc2.remove_urlpath("b.b2nd")
        blosc2.remove_urlpath("out.b2nd")


@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("fill_value", [0, 1, 0.32])
@pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all", "mean", "std", "var"])
@pytest.mark.parametrize("axis", [0, (0, 1), None])
def test_save_version3(disk, fill_value, reduce_op, axis):
    shape = (20, 50, 100)
    urlpath = "a1.b2nd" if disk else None
    if fill_value != 0:
        a = blosc2.full(shape, fill_value, urlpath=urlpath, mode="w")
        b = blosc2.full(shape, fill_value - 0.1, urlpath="b.b2nd", mode="w")
    else:
        a = blosc2.zeros(shape, dtype=np.float64, urlpath=urlpath, mode="w")
        b = blosc2.zeros(shape, dtype=np.float64, urlpath="b.b2nd", mode="w") - 0.1
    if disk:
        a = blosc2.open(urlpath)
        b = blosc2.open("b.b2nd")
    na = a[:]
    nb = b[:]

    # A reduction as a function
    expr = f"{reduce_op}(a, axis={axis}) + b"
    lexpr = blosc2.lazyexpr(expr, operands={"a": a, "b": b})
    if disk:
        lexpr.save("out.b2nd")
        lexpr = blosc2.open("out.b2nd")
    res = lexpr.compute()
    nres = getattr(na[()], reduce_op)(axis=axis) + nb
    assert np.allclose(res[()], nres)

    if disk:
        blosc2.remove_urlpath("a1.b2nd")
        blosc2.remove_urlpath("b.b2nd")
        blosc2.remove_urlpath("out.b2nd")


@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("fill_value", [0, 1, 0.32])
@pytest.mark.parametrize("reduce_op", ["sum", "prod", "min", "max", "any", "all", "mean", "std", "var"])
@pytest.mark.parametrize("axis", [0, (0, 1), None])
def test_save_version4(disk, fill_value, reduce_op, axis):
    shape = (20, 50, 100)
    urlpath = "a1.b2nd" if disk else None
    if fill_value != 0:
        a = blosc2.full(shape, fill_value, urlpath=urlpath, mode="w")
        b = blosc2.full(shape, fill_value - 0.1, urlpath="b.b2nd", mode="w")
    else:
        a = blosc2.zeros(shape, dtype=np.float64, urlpath=urlpath, mode="w")
        b = blosc2.zeros(shape, dtype=np.float64, urlpath="b.b2nd", mode="w") - 0.1
    if disk:
        a = blosc2.open(urlpath)
        b = blosc2.open("b.b2nd")
    na = a[:]

    # Just a single reduction
    expr = f"a.{reduce_op}(axis={axis})"
    lexpr = blosc2.lazyexpr(expr, operands={"a": a})
    if disk:
        lexpr.save("out.b2nd")
        lexpr = blosc2.open("out.b2nd")
    res = lexpr.compute()
    nres = getattr(na[()], reduce_op)(axis=axis)
    assert np.allclose(res[()], nres)

    if disk:
        blosc2.remove_urlpath("a1.b2nd")
        blosc2.remove_urlpath("b.b2nd")
        blosc2.remove_urlpath("out.b2nd")


@pytest.mark.parametrize("shape", [(10,), (10, 10), (10, 10, 10)])
@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("compute", [True, False])
def test_save_constructor_reduce(shape, disk, compute):
    lshape = math.prod(shape)
    urlpath_a = "a.b2nd" if disk else None
    urlpath_b = "b.b2nd" if disk else None
    a = blosc2.arange(lshape, shape=shape, urlpath=urlpath_a, mode="w")
    b = blosc2.ones(shape, urlpath=urlpath_b, mode="w")
    expr = f"arange({lshape}).sum() + a + ones({shape}).sum() + b + 1"
    lexpr = blosc2.lazyexpr(expr)
    if disk:
        lexpr.save("out.b2nd")
        lexpr = blosc2.open("out.b2nd")
    if compute:
        res = lexpr.compute()
        res = res[()]  # for later comparison with nres
    else:
        res = lexpr[()]
    na = np.arange(lshape).reshape(shape).sum()
    nb = np.ones(shape).sum()
    nres = na + a[:] + nb + b[:] + 1
    assert np.allclose(res[()], nres)
    if disk:
        blosc2.remove_urlpath(urlpath_a)
        blosc2.remove_urlpath(urlpath_b)
        blosc2.remove_urlpath("out.b2nd")


@pytest.mark.parametrize("shape", [(10,), (10, 10), (10, 10, 10)])
@pytest.mark.parametrize("disk", [True, False])
@pytest.mark.parametrize("compute", [True, False])
def test_save_constructor_reduce2(shape, disk, compute):
    lshape = math.prod(shape)
    urlpath_a = "a.b2nd" if disk else None
    urlpath_b = "b.b2nd" if disk else None
    a = blosc2.arange(lshape, shape=shape, urlpath=urlpath_a, mode="w")
    b = blosc2.ones(shape, urlpath=urlpath_b, mode="w")
    expr = "sum(a + 1) + (b + 2).sum() + 3"
    lexpr = blosc2.lazyexpr(expr)
    if disk:
        lexpr.save("out.b2nd")
        lexpr = blosc2.open("out.b2nd")
    if compute:
        res = lexpr.compute()
        res = res[()]  # for later comparison with nres
    else:
        res = lexpr[()]
    na = np.arange(lshape).reshape(shape)
    nb = np.ones(shape)
    nres = np.sum(na + 1) + (nb + 2).sum() + 3
    assert np.allclose(res, nres)
    assert res.dtype == nres.dtype
    if disk:
        blosc2.remove_urlpath(urlpath_a)
        blosc2.remove_urlpath(urlpath_b)
        blosc2.remove_urlpath("out.b2nd")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np

import blosc2

nchunks = 10
input_dtype = np.dtype(np.int64)

# Set the compression and decompression parameters
cparams = blosc2.CParams(codec=blosc2.Codec.LZ4, typesize=input_dtype.itemsize)
dparams = blosc2.DParams(nthreads=1)
contiguous = False
urlpath = None
storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath, mode="a")
# Remove previous SChunk
blosc2.remove_urlpath(urlpath)
# Create and set data
chunkshape = 20_000
data = np.zeros(chunkshape * nchunks, dtype=input_dtype)
schunk = blosc2.SChunk(
    chunksize=chunkshape * input_dtype.itemsize, data=data, cparams=cparams, dparams=dparams, storage=storage
)

out1 = np.empty(chunkshape * nchunks, dtype=input_dtype)
schunk.get_slice(0, chunkshape * nchunks, out=out1)


# Set postfilter with decorator
@schunk.postfilter(input_dtype)
def postfilter(input, output, offset):
    for i in range(input.size):
        output[i] = offset + i


out2 = np.empty(chunkshape * nchunks, dtype=input_dtype)
schunk.get_slice(0, chunkshape * nchunks, out=out2)

res = np.arange(out1.size, dtype=input_dtype)
# Check postfilter is applied
assert np.array_equal(res, out2)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

from time import time

import numexpr as ne
import numpy as np

import blosc2

shape = (4_000, 5_000)
chunks = (10, 5_000)
blocks = (1, 1000)
# Comment out the next line to force chunks and blocks above
chunks, blocks = None, None
# Check with fast compression
cparams = blosc2.CParams(clevel=1, codec=blosc2.Codec.BLOSCLZ)

print(f"*** Working with an struct array with shape: {shape}")
# Create a structured NumPy array
npa_ = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
npb_ = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
nps = np.empty(shape, dtype=[('a', npa_.dtype), ('b', npb_.dtype)])
nps['a'] = npa_
nps['b'] = npb_
npa = nps['a']
npb = nps['b']
t0 = time()
npc = npa**2 + npb**2 > 2 * npa * npb + 1
npd = np.where(npc, 0, 1)
t = time() - t0
print(f"Time to compute where expression (NumPy): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

t0 = time()
npc = ne.evaluate('where(a**2 + b**2 > 2 * a * b + 1, 0, 1)', local_dict={'a': npa, 'b': npb})
t = time() - t0
print(f"Time to compute where expression (NumExpr): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

s = blosc2.asarray(nps, chunks=chunks, blocks=blocks, cparams=cparams)
print(f"*** Working with NDArray with shape: {s.shape}, chunks: {s.chunks}, blocks: {s.blocks},"
      f" cratio: {s.schunk.cratio:.2f}x")
a = s['a']
b = s['b']

# Get a LazyExpr instance
# Compute: output is a NDArray
t0 = time()
c = a**2 + b**2 > 2 * a * b + 1
d = c.where(0, 1).compute(cparams=cparams)
t = time() - t0
print(f"Time to compute where expression (compute): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

# Compute the whole slice: output is a NumPy array
t0 = time()
c = a**2 + b**2 > 2 * a * b + 1
npd = c.where(0, 1)[:]
t = time() - t0
print(f"Time to compute where expression (getitem): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

print("*** Extracting rows")
# Compute and get row values: NumPy
t0 = time()
npc = npa**2 + npb**2 > 2 * npa * npb + 1
npd = nps[npc]
t = time() - t0
print(f"Time to get row values (NumPy): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

# Compute and get row values: output is a NDArray
t0 = time()
npd = s[a**2 + b**2 > 2 * a * b + 1].compute(cparams=cparams)
t = time() - t0
print(f"Time to get row values (compute): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

# Compute and get row values: output is a NDArray
t0 = time()
npd = s['a**2 + b**2 > 2 * a * b + 1'].compute(cparams=cparams)
t = time() - t0
print(f"Time to get row values (compute, string): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

# Compute and get row values: output is a NumPy array
t0 = time()
npd = s[a**2 + b**2 > 2 * a * b + 1][:]
t = time() - t0
print(f"Time to get row values (getitem): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")

# Compute and get row values: output is a NumPy array
t0 = time()
npd = s['a**2 + b**2 > 2 * a * b + 1'][:]
t = time() - t0
print(f"Time to get row values (getitem, string): {t:.3f} s; {nps.nbytes/2**30/t:.3f} GB/s")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import numpy as np

import blosc2

shape = (50, 50)
chunks = (49, 49)
dtype = np.float64
typesize = dtype.itemsize

# Create a NDArray from a NumPy array
random = np.random.default_rng()
array = random.normal(0, 1, np.prod(shape)).reshape(shape)
# Use ZFP_RATE codec
cparams = blosc2.CParams(codec=blosc2.Codec.ZFP_RATE, codec_meta=37)
a = blosc2.asarray(array, chunks=chunks, cparams=cparams)
print("compression ratio:", a.schunk.cratio)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
import pathlib

import numpy as np
import pytest

import blosc2

ROOT = "b2tests"
DIR = "expr/"

pytestmark = pytest.mark.network


def udf1p(inputs_tuple, output, offset):
    x = inputs_tuple[0]
    output[:] = x + 1


@pytest.mark.parametrize("chunked_eval", [True, False])
@pytest.mark.parametrize(
    ("chunks", "blocks"),
    [
        pytest.param((30, 30), (30, 30), marks=pytest.mark.heavy),
        (
            (50, 50),
            (30, 50),
        ),
    ],
)
def test_1p(chunks, blocks, chunked_eval, c2sub_context):
    dtype = np.float64
    shape = (60, 60)
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-(True, False)-a1-{shape}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    a = blosc2.C2Array(path)
    npa = a[:]
    npc = npa + 1

    expr = blosc2.lazyudf(
        udf1p, (a,), npa.dtype, chunked_eval=chunked_eval, chunks=chunks, blocks=blocks, dparams={}
    )
    res = expr.compute()
    assert res.chunks == chunks
    assert res.blocks == blocks
    assert res.dtype == npa.dtype

    tol = 1e-5 if res.dtype is np.float32 else 1e-14
    np.testing.assert_allclose(res[...], npc, rtol=tol, atol=tol)
    np.testing.assert_allclose(expr[...], npc, rtol=tol, atol=tol)


def udf2p(inputs_tuple, output, offset):
    x = inputs_tuple[0]
    y = inputs_tuple[1]
    for i in range(x.shape[0]):
        for j in range(x.shape[1]):
            output[i, j] = x[i, j] ** 2 + y[i, j] ** 2 + 2 * x[i, j] * y[i, j] + 1


@pytest.mark.parametrize("chunked_eval", [True, False])
@pytest.mark.parametrize(
    ("chunks", "blocks", "slices", "urlpath", "contiguous"),
    [
        pytest.param((53, 20), (10, 13), (slice(3, 8), slice(9, 12)), None, False),
    ],
)
def test_getitem(chunks, blocks, slices, urlpath, contiguous, chunked_eval, c2sub_context):
    dtype = np.float64
    shape = (60, 60)
    blosc2.remove_urlpath(urlpath)

    urlpath_a = f"ds-0-10-linspace-{dtype.__name__}-(True, False)-a1-{shape}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath_a}").as_posix()
    a = blosc2.C2Array(path)

    urlpath_b = f"ds-0-10-linspace-{dtype.__name__}-(False, False)-a3-{shape}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath_b}").as_posix()
    b = blosc2.C2Array(path)
    npa = a[:]
    npb = b[:]
    npc = npa**2 + npb**2 + 2 * npa * npb + 1
    dparams = {"nthreads": 4}

    expr = blosc2.lazyudf(
        udf2p,
        (npa, b),
        npa.dtype,
        chunked_eval=chunked_eval,
        chunks=chunks,
        blocks=blocks,
        storage=blosc2.Storage(urlpath=urlpath, contiguous=contiguous),
        dparams=dparams,
    )
    lazy_eval = expr[slices]
    np.testing.assert_allclose(lazy_eval, npc[slices])

    res = expr.compute(item=slices)
    np.testing.assert_allclose(res[...], npc[slices])
    assert res.schunk.urlpath is None
    assert res.schunk.contiguous == contiguous
    # Check dparams after a getitem and an eval
    assert res.schunk.dparams.nthreads == dparams["nthreads"]

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("contiguous", "urlpath", "cparams", "nchunks", "start", "stop"),
    [
        (True, None, {"typesize": 4}, 10, 0, 100),
        (True, "b2frame", {"typesize": 4}, 1, 7, 23),
        (
            False,
            None,
            {"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 5, "typesize": 4},
            5,
            21,
            200 * 2 * 100,
        ),
        (False, "b2frame", {"codec": blosc2.Codec.LZ4HC, "typesize": 4}, 7, None, None),
        (True, None, {"blocksize": 200 * 100, "typesize": 4}, 5, -2456, -234),
        (True, "b2frame", {"blocksize": 200 * 100, "typesize": 4}, 4, 2456, -234),
        (False, None, {"blocksize": 100 * 100, "typesize": 4}, 2, -200 * 100 + 234, 40000),
        (True, None, {"blocksize": 100 * 100, "typesize": 4}, 2, 0, None),
    ],
)
def test_schunk_get_slice(contiguous, urlpath, cparams, nchunks, start, stop):
    kwargs = {"contiguous": contiguous, "urlpath": urlpath, "cparams": cparams}
    schunk = blosc2.SChunk(chunksize=200 * 100 * 4, mode="w", **kwargs)
    for i in range(nchunks):
        chunk = np.full(schunk.chunksize // schunk.typesize, i, dtype=np.int32)
        schunk.append_data(chunk)

    aux = np.empty(200 * 100 * nchunks, dtype=np.int32)
    schunk.get_slice(start, stop, aux)
    if stop is None and start is not None:
        res = aux[start]
        np.array_equal(res, blosc2.get_slice_nchunks(schunk, start))
    else:
        res = aux[start:stop]
        np.array_equal(np.unique(res), blosc2.get_slice_nchunks(schunk, (start, stop)))
        # slice variant
        np.array_equal(np.unique(res), blosc2.get_slice_nchunks(schunk, slice(start, stop)))

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "chunks", "blocks", "dtype", "urlpath", "contiguous", "meta"),
    [
        ([450], [128], [25], "|S8", "frombuffer.b2nd", True, None),
        ([20, 134, 13], [3, 13, 5], [3, 10, 5], np.complex128, "frombuffer.b2nd", False, {"123": 123}),
        ([45], [12], [6], "|S4", None, True, None),
        ([30, 29], [15, 28], [5, 27], np.int16, None, False, {"2": 123, "meta2": "abcdef"}),
    ],
)
def test_buffer(shape, chunks, blocks, dtype, urlpath, contiguous, meta):
    blosc2.remove_urlpath(urlpath)

    dtype = np.dtype(dtype)
    typesize = dtype.itemsize
    size = int(np.prod(shape))
    buffer = bytes(size * typesize)
    a = blosc2.frombuffer(
        buffer,
        shape,
        chunks=chunks,
        blocks=blocks,
        dtype=dtype,
        urlpath=urlpath,
        contiguous=contiguous,
        meta=meta,
    )
    buffer2 = a.tobytes()
    assert buffer == buffer2

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("shape", "dtype"),
    [
        ([450], "|S8"),
        ([20, 134, 13], np.complex128),
        ([45], "|S4"),
        ([30, 29], np.int16),
    ],
)
def test_buffer_simple(shape, dtype):
    dtype = np.dtype(dtype)
    typesize = dtype.itemsize
    size = int(np.prod(shape))
    buffer = bytes(size * typesize)
    a = blosc2.frombuffer(buffer, shape, dtype=dtype)
    buffer2 = a.tobytes()
    assert buffer == buffer2



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

from dataclasses import asdict

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "dtype", "cparams", "urlpath", "contiguous"),
    [
        (
            (32, 18),
            np.float32,
            blosc2.CParams(codec=blosc2.Codec.NDLZ, codec_meta=4),
            None,
            False,
        ),
        (
            # For some reason, ZFP needs to always split buffers in this test
            (100, 1230),
            np.float64,
            {"codec": blosc2.Codec.ZFP_ACC, "codec_meta": 37, "splitmode": blosc2.SplitMode.ALWAYS_SPLIT},
            None,
            False,
        ),
        (
            (23, 34),
            np.float64,
            {"codec": blosc2.Codec.ZFP_PREC, "codec_meta": 37},
            "lossy.b2nd",
            True,
        ),
        (
            # For some reason, ZFP needs to always split buffers in this test
            (80, 51, 60),
            np.float32,
            {"codec": blosc2.Codec.ZFP_RATE, "codec_meta": 37, "splitmode": blosc2.SplitMode.ALWAYS_SPLIT},
            "lossy.b2nd",
            False,
        ),
        (
            (13, 13),
            np.int32,
            {"filters": [blosc2.Filter.NDMEAN], "filters_meta": [4]},
            None,
            True,
        ),
        (
            (10, 10),
            np.int64,
            {"filters": [blosc2.Filter.NDCELL], "filters_meta": [4]},
            None,
            False,
        ),
    ],
)
def test_lossy(shape, cparams, dtype, urlpath, contiguous):
    cparams_dict = cparams if isinstance(cparams, dict) else asdict(cparams)
    if cparams_dict.get("codec") == blosc2.Codec.NDLZ:
        dtype = np.uint8
    array = np.linspace(0, np.prod(shape), np.prod(shape), dtype=dtype).reshape(shape)
    a = blosc2.asarray(array, cparams=cparams, urlpath=urlpath, contiguous=contiguous, mode="w")

    if (
        a.schunk.cparams.codec in (blosc2.Codec.ZFP_RATE, blosc2.Codec.ZFP_PREC, blosc2.Codec.ZFP_ACC)
        or a.schunk.cparams.filters[0] == blosc2.Filter.NDMEAN
    ):
        _ = a[...]
    else:
        tol = 1e-5
        if dtype in (np.float32, np.float64):
            np.testing.assert_allclose(a[...], array, rtol=tol, atol=tol)
        else:
            np.array_equal(a[...], array)

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


# Example for writing and reading memory-mapped files

import numpy as np

import blosc2

urlpath = "array.b2nd"
blosc2.remove_urlpath(urlpath)
a = np.arange(1_000_000, dtype=np.int64)

# Optional: the size of the array is generous enough for the mapping size since we expect the compressed data to be
# smaller than the original size
initial_mapping_size = a.size * a.itemsize

# mmap_mode and initial_mapping_size can be used for all functions which create arrays on disk
# (SChunk, asarray, empty, etc.)
blosc2.asarray(a, urlpath=urlpath, mmap_mode="w+", initial_mapping_size=initial_mapping_size)

# Read the ndarray back via the general open function
a_read = blosc2.open(urlpath, mmap_mode="r")

assert np.all(a == a_read)
blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# This shows how to build expressions with a general mix of NDArray and NumPy operands.

import numpy as np

import blosc2

shape = (50, 50)

# Create a NDArray from a NumPy array
npa = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
npb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
npc = npa**2 + npb**2 + 2 * npa * npb + 1

a = blosc2.asarray(npa)
b = blosc2.asarray(npb)

# Get a LazyExpr instance with all NDArray operands
c = blosc2.lazyexpr("a**2 + b**2 + 2 * a * b + 1", {"a": a, "b": b})
d = c.compute()
assert np.allclose(d[:], npc)

# A LazyExpr instance with a mix of NDArray and NumPy operands
c = blosc2.lazyexpr("a**2 + b**2 + 2 * a * b + 1", {"a": npa, "b": b})
d = c.compute()
assert np.allclose(d[:], npc)

# A LazyExpr instance with a all NumPy operands
c = blosc2.lazyexpr("a**2 + b**2 + 2 * a * b + 1", {"a": npa, "b": npb})
d = c.compute()
assert np.allclose(d[:], npc)

# Evaluate partial slices
npd = c[1]
# Check
assert np.allclose(npd, npc[1])

npd = c[1:10]
# Check
assert np.allclose(npd, npc[1:10])

print(d.info)

print("Lazy expression evaluated correctly in-memory!")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
import pathlib

import numexpr as ne
import numpy as np
import pytest

import blosc2

pytestmark = pytest.mark.network

ROOT = "b2tests"
DIR = "expr/"


def get_arrays(shape, chunks_blocks):
    dtype = np.float64
    nelems = np.prod(shape)
    na1 = np.linspace(0, 10, nelems, dtype=dtype).reshape(shape)
    cleanup_paths = []
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a1-{shape}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    cleanup_paths.append(path)
    a1 = blosc2.C2Array(path)
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a2-{shape}d.b2nd"
    cleanup_paths.append(urlpath)
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    a2 = blosc2.C2Array(path)
    # Let other operands be local, on-disk NDArray copies
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a3-{shape}d.b2nd"
    cleanup_paths.append(urlpath)
    a3 = blosc2.asarray(a2, urlpath=urlpath, mode="w")
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a4-{shape}d.b2nd"
    cleanup_paths.append(urlpath)
    a4 = a3.copy(urlpath=urlpath, mode="w")
    assert isinstance(a1, blosc2.C2Array)
    assert isinstance(a2, blosc2.C2Array)
    assert isinstance(a3, blosc2.NDArray)
    assert isinstance(a4, blosc2.NDArray)

    p1 = blosc2.Proxy(a1, urlpath="p1.b2nd", mode="w")
    p3 = blosc2.Proxy(a3, urlpath="p3.b2nd", mode="w")
    cleanup_paths.extend(["p1.b2nd", "p3.b2nd"])

    return p1, a2, p3, a4, na1, np.copy(na1), np.copy(na1), np.copy(na1), cleanup_paths


@pytest.mark.parametrize(
    "chunks_blocks",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_expr_proxy_operands(chunks_blocks, c2sub_context):
    shape = (60, 60)
    a1, a2, a3, a4, na1, na2, na3, na4, cleanup_paths = get_arrays(shape, chunks_blocks)

    # Slice
    sl = slice(10)
    expr = a1 + a2 + a3 + a4
    expr += 3
    nres = ne.evaluate("na1 + na2 + na3 + na4 + 3")
    res = expr.compute(item=sl)
    np.testing.assert_allclose(res[:], nres[sl])

    # Save
    urlpath = "expr_proxies.b2nd"
    expr.save(urlpath=urlpath, mode="w")
    del expr
    expr_opened = blosc2.open("expr_proxies.b2nd")
    assert isinstance(expr_opened, blosc2.LazyExpr)

    # All
    res = expr_opened.compute()
    np.testing.assert_allclose(res[:], nres)

    # Cleanup
    blosc2.remove_urlpath(urlpath)
    for path in cleanup_paths:
        blosc2.remove_urlpath(path)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Benchmark to compute expressions with numba and NDArray instances as operands.
# As numba takes a while to compile the first time, we use cached functions, so
# make sure to run the script at least a couple of times.

from time import time

import numba as nb
import numexpr as ne
import numpy as np

import blosc2

shape = (5000, 10_000)
chunks = [500, 10_000]
blocks = [4, 10_000]
# Comment out the next line to enforce chunks and blocks above
chunks, blocks = None, None
# Check with fast compression
cparams = blosc2.CParams(clevel=1, codec=blosc2.Codec.BLOSCLZ)

dtype = np.float32
rtol = 1e-6 if dtype == np.float32 else 1e-17
atol = 1e-6 if dtype == np.float32 else 1e-17

# Expression to compute
exprs = ("x + 1",
         "x**2 + y**2 + 2 * x * y + 1",
         "sin(x)**3 + cos(y)**2 + cos(x) * sin(y) + z",
         )


# Create input arrays
npx = np.linspace(0, 1, np.prod(shape), dtype=dtype).reshape(shape)
npy = np.linspace(-1, 1, np.prod(shape), dtype=dtype).reshape(shape)
npz = np.linspace(0, 10, np.prod(shape), dtype=dtype).reshape(shape)
vardict = {"x": npx, "y": npy, "z": npz, "np": np}
x = blosc2.asarray(npx, chunks=chunks, blocks=blocks, cparams=cparams)
y = blosc2.asarray(npy, chunks=chunks, blocks=blocks, cparams=cparams)
z = blosc2.asarray(npz, chunks=chunks, blocks=blocks, cparams=cparams)
b2vardict = {"x": x, "y": y, "z": z, "blosc2": blosc2}

print(f"shape: {x.shape}, chunks: {x.chunks}, blocks: {x.blocks}, cratio: {x.schunk.cratio:.2f}")


# Define the functions to compute the expressions
# First the pure numba+numpy version
@nb.jit(parallel=True, cache=True)
def func_numba(x, y, z, n):
    output = np.empty(x.shape, x.dtype)
    if n == 0:
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(x.shape[1]):
                output[i, j] = x[i, j] + 1
    elif n == 1:
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(x.shape[1]):
                output[i, j] = x[i, j]**2 + y[i, j]**2 + 2 * x[i, j] * y[i, j] + 1
    elif n == 2:
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(x.shape[1]):
                output[i, j] = np.sin(x[i, j])**3 + np.cos(y[i, j])**2 + np.cos(x[i, j]) * np.sin(y[i, j]) + z[i, j]
    return output


# Now, the numba+blosc2 version using an udf
@nb.jit(parallel=True, cache=True)
def udf_numba(inputs, output, offset):
    icount = len(inputs)
    x = inputs[0]
    if icount == 1:
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(x.shape[1]):
                output[i, j] = x[i, j] + 1
    elif icount == 2:
        y = inputs[1]
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(x.shape[1]):
                output[i, j] = x[i, j]**2 + y[i, j]**2 + 2 * x[i, j] * y[i, j] + 1
    elif icount == 3:
        y = inputs[1]
        z = inputs[2]
        for i in nb.prange(x.shape[0]):
            for j in nb.prange(x.shape[1]):
                output[i, j] = np.sin(x[i, j])**3 + np.cos(y[i, j])**2 + np.cos(x[i, j]) * np.sin(y[i, j]) + z[i, j]


for n, expr in enumerate(exprs):
    print(f"*** Computing expression: {expr} ...")

    # Compute the expression with NumPy/numexpr
    npexpr = expr.replace("sin", "np.sin").replace("cos", "np.cos")
    t0 = time()
    npres = eval(npexpr, vardict)
    print("NumPy took %.3f s" % (time() - t0))
    # ne.set_num_threads(1)
    # nb.set_num_threads(1)  # this does not work that well; better use the NUMBA_NUM_THREADS env var
    t0 = time()
    ne.evaluate(expr, vardict, out=np.empty_like(npx))
    print("NumExpr took %.3f s" % (time() - t0))

    # Compute the expression with Blosc2
    blosc2.cparams_dflts["codec"] = blosc2.Codec.LZ4
    blosc2.cparams_dflts["clevel"] = 5
    b2expr = expr.replace("sin", "blosc2.sin").replace("cos", "blosc2.cos")
    c = eval(b2expr, b2vardict)
    t0 = time()
    d = c.compute()
    print("LazyExpr+compute took %.3f s" % (time() - t0))
    # Check
    np.testing.assert_allclose(d[:], npres, rtol=rtol, atol=atol)
    t0 = time()
    d = c[:]
    print("LazyExpr+getitem took %.3f s" % (time() - t0))
    # Check
    np.testing.assert_allclose(d[:], npres, rtol=rtol, atol=atol)

    # nb.set_num_threads(1)
    t0 = time()
    res = func_numba(npx, npy, npz, n)
    print("Numba took %.3f s" % (time() - t0))
    np.testing.assert_allclose(res, npres, rtol=rtol, atol=atol)

    inputs = (x,)
    if n == 1:
        inputs = (x, y)
    elif n == 2:
        inputs = (x, y, z)

    expr_ = blosc2.lazyudf(udf_numba, inputs, npx.dtype,
                           chunks=chunks, blocks=blocks, cparams=cparams)
    # getitem but using chunked evaluation
    t0 = time()
    res = expr_.compute()
    print("LazyUDF+compute took %.3f s" % (time() - t0))
    np.testing.assert_allclose(res[...], npres, rtol=rtol, atol=atol)
    t0 = time()
    res = expr_[:]
    print("LazyUDF+getitem took %.3f s" % (time() - t0))
    np.testing.assert_allclose(res[...], npres, rtol=rtol, atol=atol)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np

import blosc2

nchunks = 10
# Set the compression and decompression parameters
cparams = blosc2.CParams(codec=blosc2.Codec.LZ4HC, typesize=4)
dparams = blosc2.DParams()
contiguous = True
urlpath = "filename"

storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath, mode="a")
blosc2.remove_urlpath(urlpath)
numpy_meta = {b"dtype": str(np.dtype("int32"))}
test_meta = {b"lorem": 1234}
meta = {"numpy": numpy_meta, "test": test_meta}

# Create the empty SChunk
schunk = blosc2.SChunk(chunksize=200 * 1000 * 4, meta=meta, cparams=cparams, dparams=dparams)
# Append some chunks
for i in range(nchunks):
    buffer = i * np.arange(200 * 1000, dtype="int32")
    nchunks_ = schunk.append_data(buffer)
    assert nchunks_ == (i + 1)

# Decompress the second chunk in different ways
buffer = 1 * np.arange(200 * 1000, dtype="int32")
bytes_obj = buffer.tobytes()
res = schunk.decompress_chunk(1)
assert res == bytes_obj

dest = np.empty(buffer.shape, buffer.dtype)
schunk.decompress_chunk(1, dest)
assert np.array_equal(buffer, dest)

schunk.decompress_chunk(1, memoryview(dest))
assert np.array_equal(buffer, dest)

dest = bytearray(buffer)
schunk.decompress_chunk(1, dest)
assert dest == bytes_obj

# Insert a chunk in the 5th position
buffer = 10 * np.arange(200 * 1000, dtype="int32")
schunk.insert_data(5, buffer, False)

# Update a chunk compressing the data first
buffer = 11 * np.arange(200 * 1000, dtype="int32")
chunk = blosc2.compress2(buffer, cparams=cparams)
schunk.update_chunk(7, chunk)

# Delete the 4th chunk
schunk.delete_chunk(4)

# Get the compressed chunk
schunk.get_chunk(1)

# Set a slice from the SChunk
start = 5 * 200 * 1000 + 47
stop = start + 200 * 1000 + 4
val = nchunks * np.arange(stop - start, dtype="int32")
schunk[start:stop] = val

# Get the modified slice
out = np.empty(val.shape, dtype="int32")
schunk.get_slice(start, stop, out)
assert np.array_equal(val, out)

# Expand the SChunk with __setitem__
# When a part of the slice section overflows the SChunk size,
# the remaining data is appended until stop is reached
start = nchunks * 200 * 1000 - 40
stop = start + 200 * 1000
val = nchunks * np.arange(stop - start, dtype="int32")
schunk[start:stop] = val

blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# This example shows how to use the `linspace()` constructor to create a blosc2 array.

from time import time

import numpy as np

import blosc2

N = 10_000_000

# Use a constructor inside a lazy expression
print("*** Using a constructor inside a lazy expression ***")
t0 = time()
o1 = blosc2.linspace(0, 10, N, shape=(5, N // 5))
la = blosc2.lazyexpr("o1 + 1")
print(f"Build time: {time() - t0:.3f} s")
t0 = time()
for i in range(5):
    _ = la[i]
print(f"Access time: {time() - t0:.3f} s")

t0 = time()
la = (o1 + 1).sum()
print(f"Build time (sum): {time() - t0:.3f} s")
t0 = time()
print("sum:", la)
print(f"Reduction time (sum): {time() - t0:.3f} s")

# Use a constructor inside a lazy expression (string form)
print("*** Using a constructor inside a lazy expression (string form) ***")
o1 = f"linspace(0, 10, {N}, shape=(5, {N} // 5))"
t0 = time()
la = blosc2.lazyexpr(f"{o1} + 1")
print(f"Build time: {time() - t0:.3f} s")
t0 = time()
for i in range(5):
    _ = la[i]
print(f"Access time: {time() - t0:.3f} s")

t0 = time()
la = blosc2.lazyexpr(f"sum({o1} + 1)")
print(f"Build time (sum): {time() - t0:.3f} s")
t0 = time()
print("sum:", la[()])
print(f"Reduction time (sum): {time() - t0:.3f} s")

# Compare with numpy
print("*** Comparison with numpy ***")
t0 = time()
o1 = np.linspace(0, 10, N).reshape(5, N // 5) + 1
print(f"Build time: {time() - t0:.3f} s")
t0 = time()
for i in range(5):
    _ = o1[i]
print(f"Access time: {time() - t0:.3f} s")

t0 = time()
print("sum:", o1.sum())
print(f"Reduction time (sum): {time() - t0:.3f} s")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize("contiguous", [True, False])
@pytest.mark.parametrize("urlpath", [None, "b2frame"])
@pytest.mark.parametrize("mode", ["w", "a"])
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks", "start", "stop"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 4}, {}, 1, 200 * 100 * 1, 200 * 100 * 2),
        ({"typesize": 4}, {"nthreads": 4}, 1, 200 * 100 * 1 - 233, 200 * 100 * 3 + 7),
        (
            {"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 5, "typesize": 4},
            {},
            5,
            21,
            200 * 2 * 100,
        ),
        ({"codec": blosc2.Codec.LZ4HC, "typesize": 4}, {}, 7, None, None),
        ({"typesize": 4, "blocksize": 200 * 100}, {}, 7, 3, -12),
        ({"blocksize": 200 * 100, "typesize": 4}, {}, 5, -2456, -234),
        ({"blocksize": 200 * 100 + 4 * 2, "typesize": 4}, {}, 2, -1, 200 * 100 * 3 + 7),
    ],
)
def test_schunk_set_slice(contiguous, urlpath, mode, cparams, dparams, nchunks, start, stop):
    storage = {"contiguous": contiguous, "urlpath": urlpath, "mode": mode}
    blosc2.remove_urlpath(urlpath)

    data = np.arange(200 * 100 * nchunks, dtype="int32")
    schunk = blosc2.SChunk(
        chunksize=200 * 100 * 4, data=data, storage=storage, cparams=cparams, dparams=dparams
    )

    _start, _stop = start, stop
    if _start is None:
        _start = 0
    elif _start < 0:
        _start += data.size
    if _stop is None:
        _stop = data.size
    elif _stop < 0:
        _stop += data.size

    val = nchunks * np.arange(_stop - _start, dtype="int32")
    schunk[start:stop] = val

    out = np.empty(val.shape, dtype="int32")

    schunk.get_slice(_start, _stop, out)
    assert np.array_equal(val, out)

    blosc2.remove_urlpath(urlpath)


def test_schunk_set_slice_raises():
    kwargs = {"contiguous": True, "urlpath": "schunk.b2frame", "cparams": {"typesize": 4}, "dparams": {}}
    blosc2.remove_urlpath(kwargs["urlpath"])

    nchunks = 2
    data = np.arange(200 * 100 * nchunks, dtype="int32")
    blosc2.SChunk(chunksize=200 * 100 * 4, data=data, **kwargs)

    schunk = blosc2.open(kwargs["urlpath"], mode="r")
    start = 200 * 100
    stop = 200 * 100 * nchunks
    val = 3 * np.arange(start, stop, dtype="int32")

    with pytest.raises(ValueError):
        schunk[start:stop] = val

    schunk = blosc2.open(kwargs["urlpath"], mode="a")
    with pytest.raises(IndexError):
        schunk[start:stop:2] = val

    stop += 4
    with pytest.raises(ValueError):
        schunk[start:stop] = val

    start = -1
    stop = -4
    with pytest.raises(ValueError):
        schunk[start:stop] = val

    start = 200 * 100 * 2 + 1
    stop = 200 * 100 * 2 * 3
    with pytest.raises(ValueError):
        schunk[start:stop] = val

    blosc2.remove_urlpath(kwargs["urlpath"])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numexpr as ne
import numpy as np
import pytest

import blosc2

###### General expressions

# Define the parameters
test_params = [
    ((10, 100), (10, 100), "float32"),
    ((10, 100), (100,), "float64"),  # using broadcasting
]


@pytest.fixture(params=test_params)
def sample_data(request):
    shape, cshape, dtype = request.param
    # The jit decorator can work with any numpy or NDArray params in functions
    a = blosc2.linspace(0, 1, shape[0] * shape[1], dtype=dtype, shape=shape)
    b = np.linspace(1, 2, shape[0] * shape[1], dtype=dtype).reshape(shape)
    c = blosc2.linspace(-10, 10, cshape[0], dtype=dtype, shape=cshape)
    return a, b, c, shape


def test_expr(sample_data):
    a, b, c, shape = sample_data
    d_blosc2 = blosc2.evaluate("((a**3 + sin(a * 2)) < c) & (b > 0)")
    d_numexpr = ne.evaluate("((a**3 + sin(a * 2)) < c) & (b > 0)")
    np.testing.assert_equal(d_blosc2, d_numexpr)


def test_expr_out(sample_data):
    a, b, c, shape = sample_data
    # Testing with an out param
    out = blosc2.zeros(shape, dtype="bool")
    d_blosc2 = blosc2.evaluate("((a**3 + sin(a * 2)) < c) & (b > 0)", out=out)
    out2 = np.zeros(shape, dtype=np.bool_)
    d_numexpr = ne.evaluate("((a**3 + sin(a * 2)) < c) & (b > 0)", out=out2)
    np.testing.assert_equal(d_blosc2, d_numexpr)
    np.testing.assert_equal(out, out2)


def test_expr_optimization(sample_data):
    a, b, c, shape = sample_data
    d_blosc2 = blosc2.evaluate("((a**3 + sin(a * 2)) < c) & (b > 0)", optimization="none")
    d_numexpr = ne.evaluate("((a**3 + sin(a * 2)) < c) & (b > 0)", optimization="none")
    np.testing.assert_equal(d_blosc2, d_numexpr)


###### Reductions


def test_reduc(sample_data):
    a, b, c, shape = sample_data
    d_blosc2 = blosc2.evaluate("sum(((a**3 + sin(a * 2)) < c) & (b > 0), axis=1)")
    a = a[:]
    b = b[:]
    c = c[:]  # ensure that all operands are numpy arrays
    d_numpy = np.sum(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=1)
    np.testing.assert_equal(d_blosc2, d_numpy)


def test_reduc_out(sample_data):
    a, b, c, shape = sample_data
    # Testing with an out param
    out = blosc2.zeros(shape[0], dtype=np.int64)
    # Both versions below should work
    d_blosc2 = blosc2.evaluate("sum(((a**3 + sin(a * 2)) < c) & (b > 0), axis=1)", out=out)
    out2 = out[:]
    d_blosc2_ = blosc2.evaluate("sum(((a**3 + sin(a * 2)) < c) & (b > 0), axis=1, out=out2)")
    a = a[:]
    b = b[:]
    c = c[:]  # ensure that all operands are numpy arrays
    out3 = out[:]
    d_numpy = np.sum(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=1, out=out3)
    np.testing.assert_equal(d_blosc2, d_numpy)
    np.testing.assert_equal(d_blosc2_, d_numpy)
    np.testing.assert_equal(out, out2)
    np.testing.assert_equal(out, out3)



import re

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize("initial_mapping_size", [None, 1000])
def test_initial_mapping_size(tmp_path, monkeypatch, capfd, initial_mapping_size):
    monkeypatch.setenv("BLOSC_INFO", "true")
    expected_mapping_size = 2**30 if initial_mapping_size is None else initial_mapping_size
    urlpath = tmp_path / "schunk.b2frame"

    # Writing via SChunk
    storage = {"contiguous": True, "urlpath": urlpath}
    chunk_nitems = 10
    nchunks = 2
    dtype = np.dtype(np.int64)

    schunk = blosc2.SChunk(
        chunksize=chunk_nitems * dtype.itemsize,
        mmap_mode="w+",
        initial_mapping_size=initial_mapping_size,
        **storage,
    )
    for i in range(nchunks):
        buffer = i * np.arange(chunk_nitems, dtype=dtype)
        nchunks_ = schunk.append_data(buffer)
        assert nchunks_ == (i + 1)
    del schunk

    captured = capfd.readouterr()
    assert (
        re.search(
            r"Opened memory-mapped file .*schunk\.b2frame in mode w\+ with an mapping size of "
            + str(expected_mapping_size),
            captured.err,
        )
        is not None
    ), captured.err

    # Reading via open
    for mmap_mode in ["r", "r+", "c"]:
        open_mapping_size = None if mmap_mode == "r" else initial_mapping_size
        schunk_open = blosc2.open(urlpath, mmap_mode=mmap_mode, initial_mapping_size=open_mapping_size)
        for i in range(nchunks):
            buffer = i * np.arange(chunk_nitems, dtype=dtype)
            bytes_obj = buffer.tobytes()
            res = schunk_open.decompress_chunk(i)
            assert res == bytes_obj

        captured = capfd.readouterr()
        mode_mapping_size = urlpath.stat().st_size if mmap_mode == "r" else expected_mapping_size
        assert (
            re.search(
                r"Opened memory-mapped file .*schunk\.b2frame in mode "
                + re.escape(mmap_mode)
                + " with an mapping size of "
                + str(mode_mapping_size),
                captured.err,
            )
            is not None
        ), captured.err

    # Writing via asarray
    nparray = np.arange(3, dtype=np.float32)
    a = blosc2.asarray(
        nparray,
        urlpath=tmp_path / "schunk2.b2frame",
        mmap_mode="w+",
        initial_mapping_size=initial_mapping_size,
    )
    assert a == nparray
    np.testing.assert_almost_equal(a[...], nparray)

    captured = capfd.readouterr()
    assert (
        re.search(
            r"Opened memory-mapped file .*schunk2\.b2frame in mode w\+ with an mapping size of "
            + str(expected_mapping_size),
            captured.err,
        )
        is not None
    ), captured.err

    # Error handling
    with pytest.raises(ValueError, match=r"w\+ mmap_mode cannot be used to open an existing file"):
        blosc2.open(urlpath, mmap_mode="w+")

    with pytest.raises(ValueError, match="initial_mapping_size can only be used with writing modes"):
        blosc2.open(urlpath, mmap_mode="r", initial_mapping_size=100)

    with pytest.raises(ValueError, match="initial_mapping_size can only be used with mmap_mode"):
        blosc2.open(urlpath, mmap_mode=None, initial_mapping_size=100)

    with pytest.raises(ValueError, match="initial_mapping_size can only be used with writing modes"):
        blosc2.SChunk(mmap_mode="r", initial_mapping_size=100, **storage)

    with pytest.raises(ValueError, match="initial_mapping_size can only be used with mmap_mode"):
        blosc2.SChunk(mmap_mode=None, initial_mapping_size=100, **storage)

    with pytest.raises(ValueError, match="Only contiguous storage is supported"):
        blosc2.SChunk(contiguous=False, urlpath="b2frame", mmap_mode="w+")

    with pytest.raises(ValueError, match="urlpath must be set"):
        blosc2.SChunk(contiguous=True, urlpath=None, mmap_mode="w+")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np

import blosc2

nchunks = 5
input_dtype = np.dtype(np.int32)
output_dtype = np.dtype(np.float32)

# Set the compression and decompression parameters
cparams = blosc2.CParams(codec=blosc2.Codec.LZ4, typesize=4)
dparams = blosc2.DParams(nthreads=1)
contiguous = True
urlpath = None
storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath, mode="a")
# Remove previous SChunk
blosc2.remove_urlpath(urlpath)
# Create and set data
data = np.arange(200 * 1000 * nchunks, dtype=input_dtype)
schunk = blosc2.SChunk(
    chunksize=200 * 1000 * input_dtype.itemsize, data=data, cparams=cparams, dparams=dparams, storage=storage
)

out1 = np.empty(200 * 1000 * nchunks, dtype=input_dtype)
schunk.get_slice(0, 200 * 1000 * nchunks, out=out1)


# Set postfilter with decorator
@schunk.postfilter(input_dtype, output_dtype)
def postfilter(input, output, offset):
    output[:] = input - np.pi


out2 = np.empty(200 * 1000 * nchunks, dtype=output_dtype)
schunk.get_slice(0, 200 * 1000 * nchunks, out=out2)

res = np.empty(out1.shape, dtype=output_dtype)
postfilter(data, res, None)
# Check postfilter is applied
assert np.allclose(res, out2)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("contiguous", "urlpath", "chunksize", "nchunks", "start", "stop"),
    [
        (True, None, 40_000, 10, 13, 59),
        (True, "b2frame", 20_000, 5, 0, 20_000 // 4 * 5),
        (False, None, 20_000, 20, 200, 20_000 // 4 + 349),
        (False, "b2frame", 40_000, 15, 40_000 // 4, 40_000 // 4 * 2),
    ],
)
def test_schunk_proxy(contiguous, urlpath, chunksize, nchunks, start, stop):
    kwargs = {"contiguous": contiguous, "cparams": {"typesize": 4}}
    num_elem = chunksize // 4 * nchunks
    data = np.arange(num_elem, dtype="int32")
    schunk = blosc2.SChunk(chunksize=chunksize, data=data, **kwargs)
    bytes_obj = data.tobytes()
    cache = blosc2.Proxy(schunk, urlpath=urlpath, mode="w")

    cache_slice = cache[slice(start, stop)]
    assert cache_slice == bytes_obj[start * data.dtype.itemsize : stop * data.dtype.itemsize]

    cache_slice = cache.fetch(slice(start, stop))
    assert cache_slice.urlpath == urlpath
    out = np.empty(stop - start, data.dtype)
    cache_slice.get_slice(start, stop, out)
    assert np.array_equal(out, data[start:stop])

    cache_eval = cache.fetch()
    assert cache_eval.urlpath == urlpath
    out = np.empty(data.shape, data.dtype)
    cache_eval.get_slice(0, None, out)
    assert np.array_equal(out, data)

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("urlpath", "chunksize", "nchunks"),
    [
        (None, 40_000, 10),
        ("b2frame", 20_000, 5),
        (None, 20_000, 20),
        ("b2frame", 40_000, 15),
    ],
)
def test_open(urlpath, chunksize, nchunks):
    kwargs = {"urlpath": urlpath, "cparams": {"typesize": 4}}
    proxy_urlpath = "proxy.b2frame"
    blosc2.remove_urlpath(urlpath)
    num_elem = chunksize // 4 * nchunks
    data = np.arange(num_elem, dtype="int32")
    schunk = blosc2.SChunk(chunksize=chunksize, data=data, **kwargs)
    bytes_obj = data.tobytes()
    proxy = blosc2.Proxy(schunk, urlpath=proxy_urlpath, mode="w")
    del proxy
    del schunk
    if urlpath is None:
        with pytest.raises(RuntimeError):
            _ = blosc2.open(proxy_urlpath)
    else:
        proxy = blosc2.open(proxy_urlpath)
        assert proxy[0 : len(data) * 4] == bytes_obj

    blosc2.remove_urlpath(urlpath)
    blosc2.remove_urlpath(proxy_urlpath)


# Test the ProxySource class
def test_proxy_source():
    # Define an object that will be used as a source
    class Source(blosc2.ProxySource):
        def __init__(self, data):
            self._data = data
            self._nbytes = len(data) * 4
            self._typesize = 4
            self._chunksize = 20

        @property
        def nbytes(self) -> int:
            return self._nbytes

        @property
        def chunksize(self) -> int:
            return self._chunksize

        @property
        def typesize(self) -> int:
            return self._typesize

        def get_chunk(self, nchunk):
            data = self._data[nchunk * self.chunksize : (nchunk + 1) * self.chunksize]
            # Compress the data
            return blosc2.compress2(data, typesize=self._typesize)

    data = np.arange(100, dtype="int32").tobytes()
    source = Source(data)
    proxy = blosc2.Proxy(source)
    assert proxy[0:100] == data



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize("contiguous", [True, False])
@pytest.mark.parametrize("urlpath", [None, "b2frame"])
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 4}, {}, 10),
    ],
)
def test_schunk_numpy(contiguous, urlpath, cparams, dparams, nchunks):
    kwargs = {"contiguous": contiguous, "urlpath": urlpath, "cparams": cparams, "dparams": dparams}
    blosc2.remove_urlpath(urlpath)

    schunk = blosc2.SChunk(chunksize=200 * 1000 * 4, **kwargs)
    for i in range(nchunks):
        buffer = i * np.arange(200 * 1000, dtype="int32")
        nchunks_ = schunk.append_data(buffer)
        assert nchunks_ == (i + 1)

    add(schunk)
    iter(schunk)
    delete(schunk)
    clear(schunk)

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize("contiguous", [True, False])
@pytest.mark.parametrize("urlpath", [None, "b2frame"])
@pytest.mark.parametrize(
    ("nbytes", "cparams", "dparams", "nchunks"),
    [
        (136, {"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 1}, {}, 10),
    ],
)
def test_schunk(contiguous, urlpath, nbytes, cparams, dparams, nchunks):
    kwargs = {"contiguous": contiguous, "urlpath": urlpath, "cparams": cparams, "dparams": dparams}

    blosc2.remove_urlpath(urlpath)
    schunk = blosc2.SChunk(chunksize=2 * nbytes, **kwargs)
    for i in range(nchunks):
        bytes_obj = b"i " * nbytes
        nchunks_ = schunk.append_data(bytes_obj)
        assert nchunks_ == (i + 1)

    add(schunk)
    to_dict(schunk)
    iter(schunk)
    delete(schunk)
    clear(schunk)

    blosc2.remove_urlpath(urlpath)


def add(schunk):
    schunk.vlmeta["vlmeta1"] = b"val1"
    schunk.vlmeta["vlmeta2"] = "val2"
    schunk.vlmeta["vlmeta3"] = {b"lorem": 4231}
    schunk.vlmeta["vlmeta4"] = [1, 2, 3]
    schunk.vlmeta["vlmeta5"] = (1, 2, 3)

    assert schunk.vlmeta["vlmeta1"] == b"val1"
    assert schunk.vlmeta["vlmeta2"] == "val2"
    assert schunk.vlmeta["vlmeta3"] == {b"lorem": 4231}
    assert schunk.vlmeta["vlmeta4"] == [1, 2, 3]
    assert schunk.vlmeta["vlmeta5"] == (1, 2, 3)
    assert "vlmeta1" in schunk.vlmeta
    assert len(schunk.vlmeta) == 5


def to_dict(schunk):
    assert schunk.vlmeta.to_dict() == {
        b"vlmeta1": b"val1",
        b"vlmeta2": "val2",
        b"vlmeta3": {b"lorem": 4231},
        b"vlmeta4": [1, 2, 3],
        b"vlmeta5": (1, 2, 3),
    }


def delete(schunk):
    # Remove one of them
    assert "vlmeta2" in schunk.vlmeta
    del schunk.vlmeta["vlmeta2"]
    assert "vlmeta2" not in schunk.vlmeta
    assert schunk.vlmeta["vlmeta1"] == b"val1"
    assert schunk.vlmeta["vlmeta3"] == {b"lorem": 4231}
    assert schunk.vlmeta["vlmeta4"] == [1, 2, 3]
    assert schunk.vlmeta["vlmeta5"] == (1, 2, 3)
    with pytest.raises(KeyError):
        schunk.vlmeta["vlmeta2"]
    assert len(schunk.vlmeta) == 4


def iter(schunk):
    keys = ["vlmeta1", "vlmeta2", "vlmeta3", "vlmeta4", "vlmeta5"]
    for i, vlmeta in enumerate(schunk.vlmeta):
        assert vlmeta == keys[i]


def clear(schunk):
    nparray = np.arange(start=0, stop=2)
    schunk.vlmeta["vlmeta2"] = nparray.tobytes()
    assert schunk.vlmeta["vlmeta2"] == nparray.tobytes()
    assert schunk.vlmeta.__len__() == 5

    schunk.vlmeta.clear()
    assert schunk.vlmeta.__len__() == 0



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    "contiguous",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    ("shape", "chunks", "blocks", "urlpath", "dtype"),
    [
        ([634], [156], [33], "test00.b2nd", np.float64),
        ([20, 134, 13], [7, 22, 5], [3, 5, 3], "test01.b2nd", np.int32),
        ([12, 13, 14, 15, 16], [4, 6, 4, 7, 5], [2, 4, 2, 3, 3], "test02.b2nd", np.float32),
    ],
)
def test_persistency(shape, chunks, blocks, urlpath, contiguous, dtype):
    blosc2.remove_urlpath(urlpath)

    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=dtype).reshape(shape)
    _ = blosc2.asarray(nparray, chunks=chunks, blocks=blocks, urlpath=urlpath, contiguous=contiguous)
    b = blosc2.open(urlpath)

    bc = b[:]

    nparray2 = np.asarray(bc).view(dtype)
    np.testing.assert_almost_equal(nparray, nparray2)

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
from abc import ABC, abstractmethod

import numpy as np

import blosc2


class ProxyNDSource(ABC):
    """
    Base interface for NDim sources in :ref:`Proxy`.
    """

    @property
    @abstractmethod
    def shape(self) -> tuple:
        """
        The shape of the source.
        """
        pass

    @property
    @abstractmethod
    def chunks(self) -> tuple:
        """
        The chunk shape of the source.
        """
        pass

    @property
    @abstractmethod
    def blocks(self) -> tuple:
        """
        The block shape of the source.
        """
        pass

    @property
    @abstractmethod
    def dtype(self) -> np.dtype:
        """
        The dtype of the source.
        """
        pass

    @property
    def cparams(self) -> blosc2.CParams:
        """
        The compression parameters of the source.

        This property is optional and can be overridden if the source has a
        different compression configuration.
        """
        return blosc2.CParams(typesize=self.dtype.itemsize)

    @abstractmethod
    def get_chunk(self, nchunk: int) -> bytes:
        """
        Return the compressed chunk in :paramref:`self`.

        Parameters
        ----------
        nchunk: int
            The unidimensional index of the chunk to retrieve.

        Returns
        -------
        out: bytes object
            The compressed chunk.
        """
        pass

    async def aget_chunk(self, nchunk: int) -> bytes:
        """
        Return the compressed chunk in :paramref:`self` asynchronously.

        Parameters
        ----------
        nchunk: int
            The index of the chunk to retrieve.

        Returns
        -------
        out: bytes object
            The compressed chunk.

        Notes
        -----
        This method is optional, and only available if the source has an async
        `aget_chunk` method.
        """
        raise NotImplementedError(
            "aget_chunk is only available if the source has an async aget_chunk method"
        )


class ProxySource(ABC):
    """
    Base interface for sources of :ref:`Proxy` that are not NDim objects.
    """

    @property
    @abstractmethod
    def nbytes(self) -> int:
        """
        The total number of bytes in the source.
        """
        pass

    @property
    @abstractmethod
    def chunksize(self) -> tuple:
        """
        The chunksize of the source.
        """
        pass

    @property
    @abstractmethod
    def typesize(self) -> int:
        """
        The typesize of the source.
        """
        pass

    @property
    def cparams(self) -> blosc2.CParams:
        """
        The compression parameters of the source.

        This property is optional and can be overridden if the source has a
        different compression configuration.
        """
        return blosc2.CParams(typesize=self.typesize)

    @abstractmethod
    def get_chunk(self, nchunk: int) -> bytes:
        """
        Return the compressed chunk in :paramref:`self`.

        Parameters
        ----------
        nchunk: int
            The index of the chunk to retrieve.

        Returns
        -------
        out: bytes object
            The compressed chunk.
        """
        pass

    async def aget_chunk(self, nchunk: int) -> bytes:
        """
        Return the compressed chunk in :paramref:`self` asynchronously.

        Parameters
        ----------
        nchunk: int
            The index of the chunk to retrieve.

        Returns
        -------
        out: bytes object
            The compressed chunk.

        Notes
        -----
        This method is optional and only available if the source has an async
        `aget_chunk` method.
        """
        raise NotImplementedError(
            "aget_chunk is only available if the source has an async aget_chunk method"
        )


class Proxy(blosc2.Operand):
    """Proxy (with cache support) for an object following the :ref:`ProxySource` interface.

    This can be used to cache chunks of a regular data container which follows the
    :ref:`ProxySource` or :ref:`ProxyNDSource` interfaces.
    """

    def __init__(
        self, src: ProxySource or ProxyNDSource, urlpath: str | None = None, mode="a", **kwargs: dict
    ):
        """
        Create a new :ref:`Proxy` to serve as a cache to save accessed chunks locally.

        Parameters
        ----------
        src: :ref:`ProxySource` or :ref:`ProxyNDSource`
            The original container.
        urlpath: str, optional
            The urlpath where to save the container that will work as a cache.
        mode: str, optional
            "a" means read/write (create if it doesn't exist); "w" means create
            (overwrite if it exists). Default is "a".
        kwargs: dict, optional
            Keyword arguments supported:

                vlmeta: dict or None
                    A dictionary with different variable length metalayers.  One entry per metalayer:
                        key: bytes or str
                            The name of the metalayer.
                        value: object
                            The metalayer object that will be serialized using msgpack.

        """
        self.src = src
        self.urlpath = urlpath
        if kwargs is None:
            kwargs = {}
        self._cache = kwargs.pop("_cache", None)

        if self._cache is None:
            meta_val = {
                "local_abspath": None,
                "urlpath": None,
                "caterva2_env": kwargs.pop("caterva2_env", False),
            }
            container = getattr(self.src, "schunk", self.src)
            if hasattr(container, "urlpath"):
                meta_val["local_abspath"] = container.urlpath
            elif isinstance(self.src, blosc2.C2Array):
                meta_val["urlpath"] = (self.src.path, self.src.urlbase, self.src.auth_token)
            meta = {"proxy-source": meta_val}
            if hasattr(self.src, "shape"):
                self._cache = blosc2.empty(
                    self.src.shape,
                    self.src.dtype,
                    chunks=self.src.chunks,
                    blocks=self.src.blocks,
                    cparams=self.src.cparams,
                    urlpath=urlpath,
                    mode=mode,
                    meta=meta,
                )
            else:
                self._cache = blosc2.SChunk(
                    chunksize=self.src.chunksize,
                    cparams=self.src.cparams,
                    urlpath=urlpath,
                    mode=mode,
                    meta=meta,
                )
                self._cache.fill_special(self.src.nbytes // self.src.typesize, blosc2.SpecialValue.UNINIT)
        self._schunk_cache = getattr(self._cache, "schunk", self._cache)
        vlmeta = kwargs.get("vlmeta")
        if vlmeta:
            for key in vlmeta:
                self._schunk_cache.vlmeta[key] = vlmeta[key]

    def fetch(self, item: slice | list[slice] | None = None) -> blosc2.NDArray | blosc2.schunk.SChunk:
        """
        Get the container used as cache with the requested data updated.

        Parameters
        ----------
        item: slice or list of slices, optional
            If not None, only the chunks that intersect with the slices
            in items will be retrieved if they have not been already.

        Returns
        -------
        out: :ref:`NDArray` or :ref:`SChunk`
            The local container used to cache the already requested data.

        Examples
        --------
        >>> import numpy as np
        >>> import blosc2
        >>> data = np.arange(20).reshape(10, 2)
        >>> ndarray = blosc2.asarray(data)
        >>> proxy = blosc2.Proxy(ndarray)
        >>> slice_data = proxy.fetch((slice(0, 3), slice(0, 2)))
        >>> slice_data[:3, :2]
        [[0 1]
        [2 3]
        [4 5]]
        """
        if item is None:
            # Full realization
            for info in self._schunk_cache.iterchunks_info():
                if info.special != blosc2.SpecialValue.NOT_SPECIAL:
                    chunk = self.src.get_chunk(info.nchunk)
                    self._schunk_cache.update_chunk(info.nchunk, chunk)
        else:
            # Get only a slice
            nchunks = blosc2.get_slice_nchunks(self._cache, item)
            for info in self._schunk_cache.iterchunks_info():
                if info.nchunk in nchunks and info.special != blosc2.SpecialValue.NOT_SPECIAL:
                    chunk = self.src.get_chunk(info.nchunk)
                    self._schunk_cache.update_chunk(info.nchunk, chunk)

        return self._cache

    async def afetch(self, item: slice | list[slice] | None = None) -> blosc2.NDArray | blosc2.schunk.SChunk:
        """
        Retrieve the cache container with the requested data updated asynchronously.

        Parameters
        ----------
        item: slice or list of slices, optional
            If provided, only the chunks intersecting with the specified slices
            will be retrieved if they have not been already.

        Returns
        -------
        out: :ref:`NDArray` or :ref:`SChunk`
            The local container used to cache the already requested data.

        Notes
        -----
        This method is only available if the :ref:`ProxySource` or :ref:`ProxyNDSource`
        have an async `aget_chunk` method.

        Examples
        --------
        >>> import numpy as np
        >>> import blosc2
        >>> import asyncio
        >>> from blosc2 import ProxyNDSource
        >>> class MyProxySource(ProxyNDSource):
        >>>     def __init__(self, data):
        >>>         # If the next source is multidimensional, it must have the attributes:
        >>>         self.data = data
        >>>         f"Data shape: {self.shape}, Chunks: {self.chunks}"
        >>>         f"Blocks: {self.blocks}, Dtype: {self.dtype}"
        >>>     @property
        >>>     def shape(self):
        >>>         return self.data.shape
        >>>     @property
        >>>     def chunks(self):
        >>>         return self.data.chunks
        >>>     @property
        >>>     def blocks(self):
        >>>         return self.data.blocks
        >>>     @property
        >>>     def dtype(self):
        >>>         return self.data.dtype
        >>>     # This method must be present
        >>>     def get_chunk(self, nchunk):
        >>>         return self.data.get_chunk(nchunk)
        >>>     # This method is optional
        >>>     async def aget_chunk(self, nchunk):
        >>>         await asyncio.sleep(0.1) # Simulate an asynchronous operation
        >>>         return self.data.get_chunk(nchunk)
        >>> data = np.arange(20).reshape(4, 5)
        >>> chunks = [2, 5]
        >>> blocks = [1, 5]
        >>> data = blosc2.asarray(data, chunks=chunks, blocks=blocks)
        >>> source = MyProxySource(data)
        >>> proxy = blosc2.Proxy(source)
        >>> async def fetch_data():
        >>>     # Fetch a slice of the data from the proxy asynchronously
        >>>     slice_data = await proxy.afetch(slice(0, 2))
        >>>     # Note that only data fetched is shown, the rest is uninitialized
        >>>     slice_data[:]
        >>> asyncio.run(fetch_data())
        >>> # Using getitem to get a slice of the data
        >>> result = proxy[1:2, 1:3]
        >>> f"Proxy getitem: {result}"
        Data shape: (4, 5), Chunks: (2, 5)
        Blocks: (1, 5), Dtype: int64
        [[0 1 2 3 4]
        [5 6 7 8 9]
        [0 0 0 0 0]
        [0 0 0 0 0]]
        Proxy getitem: [[6 7]]
        """
        if not callable(getattr(self.src, "aget_chunk", None)):
            raise NotImplementedError("afetch is only available if the source has an aget_chunk method")
        if item is None:
            # Full realization
            for info in self._schunk_cache.iterchunks_info():
                if info.special != blosc2.SpecialValue.NOT_SPECIAL:
                    chunk = await self.src.aget_chunk(info.nchunk)
                    self._schunk_cache.update_chunk(info.nchunk, chunk)
        else:
            # Get only a slice
            nchunks = blosc2.get_slice_nchunks(self._cache, item)
            for info in self._schunk_cache.iterchunks_info():
                if info.nchunk in nchunks and info.special != blosc2.SpecialValue.NOT_SPECIAL:
                    chunk = await self.src.aget_chunk(info.nchunk)
                    self._schunk_cache.update_chunk(info.nchunk, chunk)

        return self._cache

    def __getitem__(self, item: slice | list[slice]) -> np.ndarray:
        """
        Get a slice as a numpy.ndarray using the :ref:`Proxy`.

        Parameters
        ----------
        item: slice or list of slices
            The slice of the desired data.

        Returns
        -------
        out: numpy.ndarray
            An array with the data slice.

        Examples
        --------
        >>> import numpy as np
        >>> import blosc2
        >>> data = np.arange(25).reshape(5, 5)
        >>> ndarray = blosc2.asarray(data)
        >>> proxy = blosc2.Proxy(ndarray)
        >>> proxy[0:3, 0:3]
        [[ 0  1  2]
        [ 5  6  7]
        [10 11 12]
        [20 21 22]]
        >>> proxy[2:5, 2:5]
        [[12 13 14]
        [17 18 19]
        [22 23 24]]
        """
        # Populate the cache
        self.fetch(item)
        return self._cache[item]

    @property
    def dtype(self) -> np.dtype:
        """The dtype of :paramref:`self` or None if the data is unidimensional"""
        return self._cache.dtype if isinstance(self._cache, blosc2.NDArray) else None

    @property
    def shape(self) -> tuple[int]:
        """The shape of :paramref:`self`"""
        return self._cache.shape if isinstance(self._cache, blosc2.NDArray) else len(self._cache)

    @property
    def schunk(self) -> blosc2.schunk.SChunk:
        """The :ref:`SChunk` of the cache"""
        return self._schunk_cache

    @property
    def cparams(self) -> blosc2.CParams:
        """The compression parameters of the cache"""
        return self._cache.cparams

    @property
    def info(self) -> str:
        """The info of the cache"""
        if isinstance(self._cache, blosc2.NDArray):
            return self._cache.info
        raise NotImplementedError("info is only available if the source is a NDArray")

    def __str__(self):
        return f"Proxy({self.src}, urlpath={self.urlpath})"

    @property
    def vlmeta(self) -> blosc2.schunk.vlmeta:
        """
        Get the vlmeta of the cache.

        See Also
        --------
        :py:attr:`blosc2.schunk.SChunk.vlmeta`
        """
        return self._schunk_cache.vlmeta

    @property
    def fields(self) -> dict:
        """
        Dictionary with the fields of :paramref:`self`.

        Returns
        -------
        fields: dict
            A dictionary with the fields of the :ref:`Proxy`.

        See Also
        --------
        :ref:`NDField`

        Examples
        --------
        >>> import numpy as np
        >>> import blosc2
        >>> data = np.ones(16, dtype=[('field1', 'i4'), ('field2', 'f4')]).reshape(4, 4)
        >>> ndarray = blosc2.asarray(data)
        >>> proxy = blosc2.Proxy(ndarray)
        >>>  # Get a dictionary of fields from the proxy, where each field can be accessed individually
        >>> fields_dict = proxy.fields
        >>> for field_name, field_proxy in fields_dict.items():
        >>>     print(f"Field name: {field_name}, Field data: {field_proxy}")
        Field name: field1, Field data: <blosc2.proxy.ProxyNDField object at 0x114472d20>
        Field name: field2, Field data: <blosc2.proxy.ProxyNDField object at 0x10e215be0>
        >>> fields_dict['field2'][:]
        [[1. 1. 1. 1.]
         [1. 1. 1. 1.]
         [1. 1. 1. 1.]
         [1. 1. 1. 1.]]
        """
        _fields = getattr(self._cache, "fields", None)
        if _fields is None:
            return None
        return {key: ProxyNDField(self, key) for key in _fields}


class ProxyNDField(blosc2.Operand):
    def __init__(self, proxy: Proxy, field: str):
        self.proxy = proxy
        self.field = field
        self.dtype = proxy.dtype[field]
        self.shape = proxy.shape

    def __getitem__(self, item: slice | list[slice]) -> np.ndarray:
        """
        Get a slice as a numpy.ndarray using the `field` in `proxy`.

        Parameters
        ----------
        item: slice or list of slices
            The slice of the desired data.

        Returns
        -------
        out: numpy.ndarray
            An array with the data slice.
        """
        # Get the data and return the corresponding field
        nparr = self.proxy[item]
        return nparr[self.field]


class SimpleProxy(blosc2.Operand):
    """
    Simple proxy for a NumPy array (or similar) that can be used with the Blosc2 compute engine.

    This only supports the __getitem__ method. No caching is performed.

    Examples
    --------
    >>> import numpy as np
    >>> import blosc2
    >>> a = np.arange(20, dtype=np.float32).reshape(4, 5)
    >>> proxy = blosc2.SimpleProxy(a)
    >>> proxy[1:3, 2:4]
    [[ 7.  8.]
     [12. 13.]]
    """

    def __init__(self, src, chunks: tuple | None = None, blocks: tuple | None = None):
        if not hasattr(src, "shape") or not hasattr(src, "dtype"):
            # If the source is not a NumPy array, convert it to one
            src = np.asarray(src)
        self.src = src
        self.dtype = src.dtype
        self.shape = src.shape
        # Compute reasonable values for chunks and blocks
        cparams = blosc2.CParams(clevel=0)
        self.chunks, self.blocks = blosc2.compute_chunks_blocks(
            self.shape, chunks, blocks, self.dtype, **{"cparams": cparams}
        )

    def __getitem__(self, item: slice | list[slice]) -> np.ndarray:
        """
        Get a slice as a numpy.ndarray using the :ref:`ProxyNumPy`.

        Parameters
        ----------
        item

        Returns
        -------
        out: numpy.ndarray
            An array with the data slice.
        """
        return self.src[item]


def jit(func=None, *, out=None, **kwargs):  # noqa: C901
    """
    Prepare a function so that it can be used with the Blosc2 compute engine.

    The inputs of the function can be any combination of NumPy/NDArray arrays
    and scalars.  The function will be called with the NumPy arrays replaced by
    :ref:`SimpleProxy` objects, whereas NDArray objects will be used as is.

    The returned value will be a NumPy array if all arguments are NumPy arrays
    or if not kwargs are provided. Else, the return value will be a NDArray
    created using the provided kwargs.

    Parameters
    ----------
    func: callable
        The function to be prepared for the Blosc2 compute engine.
    out: np.ndarray, NDArray, optional
        The output array where the result will be stored.
    **kwargs: dict, optional
        Additional keyword arguments supported by the :func:`empty` constructor.

    Returns
    -------
    wrapper

    Notes
    -----
    * Although many NumPy functions are supported, some may not be implemented yet.
      If you find a function that is not supported, please open an issue.
    * `out` and `kwargs` parameters are not supported for all expressions
      (e.g. when using a reduction as the last function).  In this case, you can
      still use the `out` parameter of the reduction function for some custom
      control over the output.

    Examples
    --------
    >>> import numpy as np
    >>> import blosc2
    >>> @blosc2.jit
    >>> def compute_expression(a, b, c):
    >>>     return np.sum(((a ** 3 + np.sin(a * 2)) > 2 * c) & (b > 0), axis=1)
    >>> a = np.arange(20, dtype=np.float32).reshape(4, 5)
    >>> b = np.arange(20).reshape(4, 5)
    >>> c = np.arange(5)
    >>> compute_expression(a, b, c)
    [5 5 5 5]
    """

    def decorator(func):  # noqa: C901
        def wrapper(*args, **func_kwargs):
            # Get some kwargs in decorator for SimpleProxy constructor
            proxy_kwargs = {"chunks": kwargs.get("chunks"), "blocks": kwargs.get("blocks")}

            # Wrap the arguments in SimpleProxy objects if they are not NDArrays
            new_args = []
            for arg in args:
                if issubclass(type(arg), blosc2.Operand):
                    new_args.append(arg)
                else:
                    new_args.append(SimpleProxy(arg, **proxy_kwargs))
            # The same for the keyword arguments
            for key, value in func_kwargs.items():
                if issubclass(type(value), blosc2.Operand):
                    continue
                func_kwargs[key] = SimpleProxy(value, **proxy_kwargs)

            # Call function with the new arguments
            retval = func(*new_args, **func_kwargs)

            # Treat return value
            # If it is a numpy array, return it as is
            if isinstance(retval, np.ndarray):
                if kwargs and any(kwargs[key] is not None for key in kwargs):
                    # But if kwargs are provided, return a NDArray instead
                    return blosc2.asarray(retval, **kwargs)
                return retval

            # In some instances, the return value is not a LazyExpr
            # (e.g. using a reduction as the last function, and using an `out` param)
            if not isinstance(retval, blosc2.LazyExpr):
                return retval

            # If the return value is a LazyExpr, compute it
            if out is not None:
                return retval.compute(out=out, **kwargs)
            if kwargs and any(kwargs[key] is not None for key in kwargs):
                return retval.compute(**kwargs)
            # If no kwargs are provided, return a numpy array
            return retval[()]

        return wrapper

    if func is None:
        return decorator
    else:
        return decorator(func)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import sys
from time import time

import numpy as np

import blosc2

# Dimensions, type and persistence properties for the arrays
nelem = 1_00_000_000
dtype = np.dtype(np.float64)
print(f"Filling a SChunk with {nelem / 1e6} Melements of {dtype=}")

persistent = bool(sys.argv[1]) if len(sys.argv) > 1 else False
if persistent:
    urlpath = "bench_fill_special.b2frame"
    print(f"Writing output to {urlpath}...")
else:
    urlpath = None


def create_schunk(data=None):
    blosc2.remove_urlpath(urlpath)
    # Create the empty SChunk
    return blosc2.SChunk(data=data, urlpath=urlpath, cparams={"typesize": dtype.itemsize})


t0 = time()
schunk = create_schunk(data=np.full(nelem, np.pi, dtype))
t = (time() - t0) * 1000.
print(f"Time with `data` argument in constructor: {t:19.3f} ms")

schunk = create_schunk()
t0 = time()
schunk.fill_special(nelem, blosc2.SpecialValue.UNINIT)
schunk[:] = np.full(nelem, np.pi, dtype)
t = (time() - t0) * 1000.
print(f"Time without passing directly the value: {t:20.3f} ms")

schunk = create_schunk()
t0 = time()
schunk.fill_special(nelem, blosc2.SpecialValue.VALUE, np.pi)
t = (time() - t0) * 1000.
print(f"Time passing directly the value to `fill_special`: {t:10.3f} ms")

blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Show how getitem / setitem works for an NDArray

import numpy as np

import blosc2

shape = (10, 10)
slices = (slice(2, 7), slice(4, 8))

# Create a NDArray from a numpy array
nparray = np.arange(int(np.prod(shape)), dtype=np.int32).reshape(shape)
a = blosc2.asarray(nparray)

# Get a slice
buffer = a[slices]

# Set a slice
a[slices] = np.ones_like(buffer) - buffer
print(a[...])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Import structured arrays using the array interface

import numpy as np

import blosc2

shape = (2, 2)
dtype = np.float64

# Create a structured array
arr0 = np.arange(np.prod(shape), dtype=dtype).reshape(shape)
arr1 = np.arange(np.prod(shape), dtype=dtype).reshape(shape)
arr = np.array([arr0, arr1], dtype="f8,f8")
print("NumPy struct array:\n", arr)

# And convert it into a NDArray using the array interface
a = blosc2.asarray(arr)
print("\nNDArray struct array:\n", a[...])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Shows how you can persist an array on disk

import numpy as np

import blosc2

shape = (128, 128)
urlpath = "ex_persistency.b2nd"
dtype = np.complex128

# Create a NDArray from a numpy array (and save it on disk)
nparray = np.arange(int(np.prod(shape)), dtype=dtype).reshape(shape)
a = blosc2.asarray(nparray, urlpath=urlpath, mode="w")

# Read the array from disk
b = blosc2.open(urlpath)
# And see its contents
print(b[...])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import pathlib

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("mode", "mmap_mode"), [("r", None), ("w", None), ("a", None), ("r", "r"), ("w", "w+")]
)
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks"),
    [
        ({"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 5, "typesize": 4}, {}, 5),
    ],
)
def test_schunk_pathlib(mode, mmap_mode, cparams, dparams, nchunks):
    urlpath = pathlib.Path("b2frame")
    kwargs = {"urlpath": urlpath, "cparams": cparams, "dparams": dparams}
    blosc2.remove_urlpath(urlpath)

    if mode != "r":
        chunk_len = 200 * 1000
        schunk = blosc2.SChunk(chunksize=chunk_len * 4, mode=mode, mmap_mode=mmap_mode, **kwargs)
        assert schunk.urlpath == str(urlpath)

        for i in range(nchunks):
            buffer = i * np.arange(chunk_len, dtype="int32")
            nchunks_ = schunk.append_data(buffer)
            assert nchunks_ == (i + 1)

        for i in range(nchunks):
            buffer = i * np.arange(chunk_len, dtype="int32")
            dest = np.empty(buffer.shape, buffer.dtype)
            schunk.decompress_chunk(i, dest)
            assert np.array_equal(buffer, dest)

    blosc2.remove_urlpath(urlpath)


argnames = "shape, chunks, blocks, slices, dtype"
argvalues = [
    ([12, 13, 14, 15, 16], [5, 5, 5, 5, 5], [2, 2, 2, 2, 2], (slice(1, 3), ..., slice(3, 6)), np.float32),
]


@pytest.mark.parametrize(("mode", "mmap_mode"), [("w", None), (None, "w+")])
@pytest.mark.parametrize(argnames, argvalues)
def test_ndarray_pathlib(tmp_path, mode, mmap_mode, shape, chunks, blocks, slices, dtype):
    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=dtype).reshape(shape)
    a = blosc2.asarray(
        nparray, chunks=chunks, blocks=blocks, urlpath=tmp_path / "test.b2nd", mode=mode, mmap_mode=mmap_mode
    )
    b = a.slice(slices)
    np_slice = a[slices]
    assert b.shape == np_slice.shape
    np.testing.assert_almost_equal(b[...], np_slice)

    b = blosc2.open(
        tmp_path / "test.b2nd",
        mode="a" if mmap_mode is None else None,
        mmap_mode="r+" if mode is None else None,
    )
    np.testing.assert_almost_equal(b[...], nparray)

    a = blosc2.zeros(shape, dtype, urlpath=tmp_path / "test2.b2nd", mode=mode, mmap_mode=mmap_mode)
    b = np.zeros(shape, dtype)
    np.testing.assert_almost_equal(b[...], a[...])

    a = blosc2.full(shape, 3, urlpath=tmp_path / "test3.b2nd", mode=mode, mmap_mode=mmap_mode)
    b = np.full(shape, 3)
    np.testing.assert_almost_equal(b[...], a[...])

    a = blosc2.frombuffer(
        bytes(nparray), shape, dtype, urlpath=tmp_path / "test4.b2nd", mode=mode, mmap_mode=mmap_mode
    )
    np.testing.assert_almost_equal(nparray[...], a[...])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import array

# Compress and decompress different arrays
import blosc2

a = array.array("i", range(1000 * 1000))
a_bytesobj = a.tobytes()
c_bytesobj = blosc2.compress(a_bytesobj, typesize=4)
assert len(c_bytesobj) < len(a_bytesobj)
a_bytesobj2 = blosc2.decompress(c_bytesobj)
assert a_bytesobj == a_bytesobj2

dest = blosc2.compress(b"", 1)
assert blosc2.decompress(dest) == b""
assert type(blosc2.decompress(blosc2.compress(b"1" * 7, 1), as_bytearray=True)) is bytearray



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import blosc2

print(blosc2.set_releasegil(True))
print(blosc2.set_releasegil(True))



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Shows how you can make a proxy of a remote array (served with Caterva2) on disk
# Note that, for running this example, you will need the blosc2-grok package.

import os
from time import time

import blosc2

urlbase = "https://demo.caterva2.net/"
path = "example/lung-jpeg2000_10x.b2nd"
a = blosc2.C2Array(path, urlbase=urlbase)
b = blosc2.Proxy(a, urlpath="proxy.b2nd", mode="w")

# Check metadata (note that all should be the same)
print("*** Metadata ***")
print(f"Codec in 'a': {a.cparams.codec}")
print(f"Codec in 'b': {b.cparams.codec}")
print(f"Filters in 'a': {a.cparams.filters}")
print(f"Filters in 'b': {b.cparams.filters}")

# Check array properties
print("*** Array properties ***")
print(f"Shape in 'a': {a.shape}")
print(f"Shape in 'b': {b.shape}")
print(f"Type in 'a': {a.dtype}")
print(f"Type in 'b': {b.dtype}")

print("*** Fetching data ***")
t0 = time()
print(f"Data in 'a': {a[0, 0, 0:10]}")
print(f"Time to fetch data in 'a': {time() - t0:.3f}s")
t0 = time()
print(f"Data in 'b': {b[0, 0, 0:10]}")
print(f"Time to fetch data in 'b': {time() - t0:.3f}s")
t0 = time()
print(f"Data in 'b': {b[0, 0, 0:10]}")
print(f"Time to fetch data in 'b' (cached): {time() - t0:.3f}s")

# Check sizes. Note that the proxy will only have the 'touched' chunks (only 1 in this case)
print("*** Sizes ***")
print(f"Size in 'a': {a.meta['schunk']['cbytes']}")
print(f"Size in 'b': {b.schunk.cbytes}")
# Check sizes on disk
print("*** Disk sizes ***")
print(f"Size 'b' (disk): {os.stat(b.urlpath).st_size}")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

from time import time

import numpy as np

import blosc2

# Size and dtype of super-chunks
nchunks = 10_000
chunkshape = 200_000
dtype = np.dtype(np.int32)

# Set the compression and decompression parameters
dparams = {"nthreads" : 1}

# Create array
arr = blosc2.empty(shape=(nchunks * chunkshape,), chunks=(chunkshape,), dtype=dtype, dparams=dparams)
data = np.arange(chunkshape, dtype=dtype)

t0 = time()
for i in range(nchunks):
    arr[i * chunkshape : (i + 1) * chunkshape] = data
t = time() - t0
print(
    f"time append: {t:.2f}s ({arr.schunk.nbytes / (t * 2**30):.3f} GB/s)"
    f" / cratio: {arr.schunk.cratio:.2f}x"
)

t0 = time()
arr_ = arr.copy()
t = time() - t0
print(
    f"time copy (no postfilter): {t:.2f}s ({arr_.schunk.nbytes / (t * 2**30):.3f} GB/s)"
    f" / cratio: {arr_.schunk.cratio:.2f}x"
)


# Associate a postfilter to schunk
@arr.schunk.postfilter(dtype)
def py_postfilter(input, output, offset):
    output[:] = 0


t0 = time()
arr_ = arr.copy()
t = time() - t0
print(
    f"time sum (postfilter): {t:.2f}s ({arr_.schunk.nbytes / (t * 2**30):.3f} GB/s)"
    f" / cratio: {arr_.schunk.cratio:.2f}x"
)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Storing data in sparse vs contiguous mode

from time import time

import numpy as np

import blosc2

urlpath_sparse = "ex_formats_sparse.b2nd"
urlpath_contiguous = "ex_formats_contiguous.b2nd"

shape = (1000 * 1000,)
chunks = (1000,)
blocks = (100,)
dtype = np.dtype(np.float64)

t0 = time()
a = blosc2.empty(
    shape,
    dtype=dtype,
    chunks=chunks,
    blocks=blocks,
    urlpath=urlpath_sparse,
    contiguous=False,
    mode="w",
)
for nchunk in range(a.schunk.nchunks):
    a[nchunk * chunks[0] : (nchunk + 1) * chunks[0]] = np.arange(chunks[0], dtype=dtype)
t1 = time()

print(f"Time: {(t1 - t0):.4f} s")
an = a[...]

t0 = time()
b = blosc2.empty(
    shape,
    dtype=dtype,
    chunks=chunks,
    blocks=blocks,
    urlpath=urlpath_contiguous,
    contiguous=True,
    mode="w",
)

for nchunk in range(shape[0] // chunks[0]):
    b[nchunk * chunks[0] : (nchunk + 1) * chunks[0]] = np.arange(chunks[0], dtype=dtype)
t1 = time()

print(f"Time: {(t1 - t0):.4f} s")
bn = b[...]

np.testing.assert_allclose(an, bn)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import os

import numpy as np
import pytest

import blosc2

##### pack / unpack  #####


@pytest.mark.parametrize(
    ("size", "dtype"),
    [
        (1e6, "int64"),
        (1e6, "f8"),
        (1e6, "i1"),
    ],
)
def test_pack_array(size, dtype):
    nparray = np.arange(size, dtype=dtype)
    parray = blosc2.pack_array(nparray)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert len(parray) < nparray.size * nparray.itemsize

    a2 = blosc2.unpack_array(parray)
    assert np.array_equal(nparray, a2)


@pytest.mark.parametrize(
    ("size", "dtype"),
    [
        (1e6, "int64"),
        (1e6, "float64"),
        (1e6, np.float64),
        (1e6, np.int8),
        pytest.param(3e8, "int64", marks=pytest.mark.heavy),  # > 2 GB
    ],
)
def test_pack_array2(size, dtype):
    nparray = np.arange(size, dtype=dtype)
    parray = blosc2.pack_array2(nparray)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert len(parray) < nparray.size * nparray.itemsize

    a2 = blosc2.unpack_array2(parray)
    assert np.array_equal(nparray, a2)


@pytest.mark.parametrize(("size", "dtype"), [(100_000, "i4,i4"), (10_000, "i4,f8"), (3000, "i4,f4,S8")])
def test_pack_array2_struct(size, dtype):
    nparray = np.fromiter(iter(range(size)), dtype="i4,f4,S8")
    parray = blosc2.pack_array2(nparray)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert len(parray) < nparray.size * nparray.itemsize

    a2 = blosc2.unpack_array2(parray)
    assert np.array_equal(nparray, a2)


@pytest.mark.parametrize(
    ("size", "dtype"),
    [
        (1e6, "float32"),
        (1e6, "float64"),
        (1e6, "int8"),
    ],
)
def test_pack_tensor_torch(size, dtype):
    torch = pytest.importorskip("torch")
    dtype = getattr(torch, dtype)
    tensor = torch.arange(size, dtype=dtype)
    cframe = blosc2.pack_tensor(tensor)
    atensor = np.asarray(tensor)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert len(cframe) < atensor.size * atensor.dtype.itemsize

    tensor2 = blosc2.unpack_tensor(cframe)
    assert np.array_equal(atensor, np.asarray(tensor2))


@pytest.mark.parametrize(
    ("size", "dtype"),
    [
        (1e6, np.float32),
        (1e6, np.float64),
        (1e6, np.int8),
    ],
)
def test_pack_tensor_tensorflow(size, dtype):
    tensorflow = pytest.importorskip("tensorflow")
    array = np.arange(size, dtype=dtype)
    tensor = tensorflow.constant(array)
    cframe = blosc2.pack_tensor(tensor)
    atensor = np.asarray(tensor)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert len(cframe) < atensor.size * atensor.dtype.itemsize

    tensor2 = blosc2.unpack_tensor(cframe)
    assert np.array_equal(atensor, np.asarray(tensor2))


@pytest.mark.parametrize(
    ("size", "dtype"),
    [
        (1e6, "int64"),
        (1e6, "float64"),
        (1e6, np.float64),
        (1e6, np.int8),
        pytest.param(3e8, "int64", marks=pytest.mark.heavy),  # > 2 GB
    ],
)
def test_pack_tensor_array(size, dtype):
    nparray = np.arange(size, dtype=dtype)
    parray = blosc2.pack_tensor(nparray)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert len(parray) < nparray.size * nparray.itemsize

    a2 = blosc2.unpack_tensor(parray)
    assert np.array_equal(nparray, a2)


def test_pack_tensor_empty():
    empty = np.zeros((0,), dtype=float)
    pempty = blosc2.pack_tensor(empty)

    empty2 = blosc2.unpack_tensor(pempty)
    assert np.array_equal(empty, empty2)
    assert empty2.dtype == empty.dtype
    assert empty2.shape == empty.shape


##### save / load  #####


@pytest.mark.parametrize(
    ("size", "dtype", "urlpath"),
    [
        (1e6, "int64", "test.bl2"),
        (1e6, "float32", "test.bl2"),
    ],
)
def test_save_array(size, dtype, urlpath):
    nparray = np.arange(size, dtype=dtype)
    serial_size = blosc2.save_array(nparray, urlpath, mode="w")
    if not os.getenv("BTUNE_TRADEOFF"):
        assert serial_size < nparray.size * nparray.itemsize

    a2 = blosc2.load_array(urlpath)
    blosc2.remove_urlpath(urlpath)
    assert np.array_equal(nparray, a2)


@pytest.mark.parametrize(
    ("size", "dtype", "urlpath"),
    [
        (1e6, "int64", "test.bl2"),
        (1e6, "float32", "test.bl2"),
    ],
)
def test_save_tensor_array(size, dtype, urlpath):
    nparray = np.arange(size, dtype=dtype)
    serial_size = blosc2.save_tensor(nparray, urlpath, mode="w")
    if not os.getenv("BTUNE_TRADEOFF"):
        assert serial_size < nparray.size * nparray.itemsize

    a2 = blosc2.load_tensor(urlpath)
    blosc2.remove_urlpath(urlpath)
    assert np.array_equal(nparray, a2)


@pytest.mark.parametrize(
    ("size", "dtype", "urlpath"),
    [
        (1e6, "int64", "test.bl2"),
        (1e6, "float32", "test.bl2"),
    ],
)
def test_save_tensor_tensorflow(size, dtype, urlpath):
    tensorflow = pytest.importorskip("tensorflow")
    nparray = np.arange(size, dtype=dtype)
    tensor = tensorflow.constant(nparray)
    serial_size = blosc2.save_tensor(tensor, urlpath, mode="w")
    if not os.getenv("BTUNE_TRADEOFF"):
        assert serial_size < nparray.size * nparray.itemsize

    tensor2 = blosc2.load_tensor(urlpath)
    blosc2.remove_urlpath(urlpath)
    assert np.array_equal(nparray, np.asarray(tensor2))


@pytest.mark.parametrize(
    ("size", "dtype", "urlpath"),
    [
        (1e6, "int64", "test.bl2"),
        (1e6, "float32", "test.bl2"),
    ],
)
def test_save_tensor_torch(size, dtype, urlpath):
    torch = pytest.importorskip("torch")
    nparray = np.arange(size, dtype=dtype)
    tensor = torch.tensor(nparray)
    serial_size = blosc2.save_tensor(tensor, urlpath, mode="w")
    if not os.getenv("BTUNE_TRADEOFF"):
        assert serial_size < nparray.size * nparray.itemsize

    tensor2 = blosc2.load_tensor(urlpath)
    blosc2.remove_urlpath(urlpath)
    assert np.array_equal(nparray, np.asarray(tensor2))


@pytest.mark.parametrize(
    ("size", "sparse", "urlpath"),
    [
        (1e6, True, "test.bl2"),
        (1e6, False, "test.bl2"),
    ],
)
def test_save_tensor_sparse(size, sparse, urlpath):
    nparray = np.arange(size, dtype=np.int32)
    serial_size = blosc2.save_tensor(nparray, urlpath, mode="w", contiguous=not sparse)
    if not os.getenv("BTUNE_TRADEOFF"):
        assert serial_size < nparray.size * nparray.itemsize

    a2 = blosc2.load_tensor(urlpath)
    assert os.path.isdir(urlpath) == sparse
    blosc2.remove_urlpath(urlpath)
    assert np.array_equal(nparray, a2)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Small benchmark for computing outer products using the broadcast feature

from time import time

import numpy as np

import blosc2

N = 10_000
# N = 1_000
# chunks = 11
# blocks = 9
shape1, shape2 = (N, 1), (N,)

# Create a NDArray from a NumPy array
npa = np.arange(np.prod(shape1), dtype=np.int64).reshape(shape1)
npb = np.arange(np.prod(shape2), dtype=np.int64).reshape(shape2)
# a = blosc2.asarray(npa, chunks=(chunks, 1), blocks=(blocks, 1))
# b = blosc2.asarray(npb, chunks=chunks, blocks=blocks)
a = blosc2.asarray(npa)
b = blosc2.asarray(npb)

for codec in blosc2.Codec:
    if codec.value > blosc2.Codec.ZSTD.value:
        break
    print(f"Codec: {codec}")
    t0 = time()
    c = a * b
    # print(f"Elapsed time (expr): {time() - t0:.6f} s")
    t0 = time()
    # d = c.compute(cparams=dict(codec=codec, clevel=5), chunks=(chunks, chunks), blocks=(blocks, blocks))
    d = c.compute(cparams={"codec": codec, "clevel": 5})
    print(f"Elapsed time (compute): {time() - t0:.2f}s")
    # print(d[:])
    print(f"cratio: {d.schunk.cratio:.2f}x")
    # print(d.info)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Filter and sort fields in a structured array
# Note that this only works for 1D arrays

import sys
from time import time

import numpy as np

import blosc2

N = 1_000_000

# arr = blosc2.open("/Users/faltet/Downloads/ds-1d-fields.b2nd")
# Create a numpy structured array with 3 fields and N elements
dt = np.dtype([("a", "i4"), ("b", "f4"), ("c", "f8")])
nsa = np.empty((N,), dtype=dt)
# TODO: Make this work with a 2D array
# nsa = np.empty((N,N), dtype=dt)
nsa["a"][:] = np.arange(N, dtype="i4")
nsa["b"][:] = np.linspace(0, 1, N, dtype="f4")
rng = np.random.default_rng(42)  # to get reproducible results
nsa["c"][:] = rng.random(N)

arr = blosc2.asarray(nsa)

t0 = time()
# Using plain sort in combination with filter
# farr = arr["b >= c"].sort("c").compute()
# You can use indices() to get the indices sorted
farr = arr["b >= c"].indices(order="c").compute()
# You can also use __getitem__ to get numpy arrays as result
# farr = arr["b >= c"].sort("c")[:]
print(f"Time to filter: {time() - t0:.3f} s")
print(f"farr: {farr[:10]}")
if farr.dtype == np.dtype("int64"):
    print(f"sorted (blosc2):\n {arr[farr[:10]]}")

print(f"len(farr): {len(farr)}, len(arr): {len(arr)}")
print(f"type of farr: {farr.dtype}, type of arr: {arr.dtype}")

if isinstance(farr, np.ndarray):
    print(f"nbytes of farr: {farr.nbytes / 2**20:.2f}MB")
    # We cannot proceed anymore
    sys.exit(1)

print(f"cratio of farr: {farr.schunk.cratio:.2f}, cratio of arr: {arr.schunk.cratio:.2f}")
print(
    f"nbytes of farr: {farr.schunk.nbytes / 2**20:.2f}MB, nbytes of arr: {arr.schunk.nbytes / 2**20:.2f}MB"
)
print(
    f"cbytes of farr: {farr.schunk.cbytes / 2**20:.2f}MB, cbytes of arr: {arr.schunk.cbytes / 2**20:.2f}MB"
)
print(f"cparams of farr: {farr.cparams}, cparams of arr: {arr.cparams}")
print(f"chunks of farr: {farr.chunks}, chunks of arr: {arr.chunks}")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


# Test the python-blosc API

import ctypes
import gc
import os
import unittest

import pytest

import blosc2

try:
    import numpy as np
except ImportError:
    has_numpy = False
else:
    has_numpy = True

try:
    import psutil
except ImportError:
    psutil = None


class TestCodec(unittest.TestCase):
    def setUp(self):
        self.PY_27_INPUT = (
            b"\x02\x01\x03\x02\x85\x00\x00\x00\x84\x00\x00"
            b"\x00\x95\x00\x00\x00\x80\x02cnumpy.core.multiarray"
            b"\n_reconstruct\nq\x01cnumpy\nndarray\nq\x02K\x00\x85U"
            b"\x01b\x87Rq\x03(K\x01K\x05\x85cnumpy\ndtype\nq\x04U\x02S2K"
            b"\x00K\x01\x87Rq\x05(K\x03U\x01|NNNK\x02K\x01K\x00tb\x89U\n\xc3"
            b"\xa5\xc3\xa7\xc3\xb8\xcf\x80\xcb\x9atb."
        )

    def test_basic_codec(self):
        s = b"0123456789"
        c = blosc2.compress(s, typesize=1)
        d = blosc2.decompress(c)
        assert s == d

    def test_all_compressors(self):
        s = b"0123456789" * 100
        for codec in blosc2.compressor_list():
            c = blosc2.compress(s, typesize=1, codec=codec)
            d = blosc2.decompress(c)
            assert s == d

    def test_all_filters(self):
        s = b"0123456789" * 100
        filters = list(blosc2.Filter)
        for filter_ in filters:
            c = blosc2.compress(s, typesize=1, filter=filter_)
            d = blosc2.decompress(c)
            assert s == d

    def test_set_nthreads_exceptions(self):
        with pytest.raises(ValueError):
            blosc2.set_nthreads(2**31)

    def test_compress_input_types(self):
        import numpy as np

        # assume the expected answer was compressed from bytes
        expected = blosc2.compress(b"0123456789", typesize=1)

        # now for all the things that support the buffer interface
        assert expected == blosc2.compress(memoryview(b"0123456789"), typesize=1)

        assert expected == blosc2.compress(bytearray(b"0123456789"), typesize=1)
        assert expected == blosc2.compress(np.array([b"0123456789"]), typesize=1)

    def test_decompress_input_types(self):
        import numpy as np

        # assume the expected answer was compressed from bytes
        expected = b"0123456789"
        compressed = blosc2.compress(expected, typesize=1)

        # now for all the things that support the buffer interface
        assert expected == blosc2.decompress(compressed)
        assert expected == blosc2.decompress(memoryview(compressed))

        assert expected == blosc2.decompress(bytearray(compressed))
        assert expected == blosc2.decompress(np.array([compressed]))

    def test_decompress_releasegil(self):
        import numpy as np

        # assume the expected answer was compressed from bytes
        blosc2.set_releasegil(True)
        expected = b"0123456789"
        compressed = blosc2.compress(expected, typesize=1)

        # now for all the things that support the buffer interface
        assert expected == blosc2.decompress(compressed)
        assert expected == blosc2.decompress(memoryview(compressed))

        assert expected == blosc2.decompress(bytearray(compressed))
        assert expected == blosc2.decompress(np.array([compressed]))
        blosc2.set_releasegil(False)

    def test_decompress_input_types_as_bytearray(self):
        import numpy as np

        # assume the expected answer was compressed from bytes
        expected = bytearray(b"0123456789")
        compressed = blosc2.compress(expected, typesize=1)

        # now for all the things that support the buffer interface
        assert expected == blosc2.decompress(compressed, as_bytearray=True)
        assert expected == blosc2.decompress(memoryview(compressed), as_bytearray=True)

        assert expected == blosc2.decompress(bytearray(compressed), as_bytearray=True)
        assert expected == blosc2.decompress(np.array([compressed]), as_bytearray=True)

    def test_compress_exceptions(self):
        s = b"0123456789"

        with pytest.raises(ValueError):
            blosc2.compress(s, typesize=0)
        with pytest.raises(ValueError):
            blosc2.compress(s, typesize=blosc2.MAX_TYPESIZE + 1)

        with pytest.raises(ValueError):
            blosc2.compress(s, typesize=1, clevel=-1)
        with pytest.raises(ValueError):
            blosc2.compress(s, typesize=1, clevel=10)

        with pytest.raises(TypeError):
            blosc2.compress(1.0, 1)
        with pytest.raises(TypeError):
            blosc2.compress(["abc"], 1)

        # Create a simple mock to avoid having to create a buffer of 2 GB
        class LenMock:
            def __len__(self):
                return blosc2.MAX_BUFFERSIZE + 1

        with pytest.raises(ValueError):
            blosc2.compress(LenMock(), typesize=1)

    def test_decompress_exceptions(self):
        with pytest.raises(TypeError):
            blosc2.decompress(1.0)
        with pytest.raises(TypeError):
            blosc2.decompress(["abc"])

    @unittest.skipIf(not has_numpy, "Numpy not available")
    def test_pack_array_exceptions(self):
        with pytest.raises(AttributeError):
            blosc2.pack_array("abc")
        with pytest.raises(AttributeError):
            blosc2.pack_array(1.0)

        # items = (blosc2.MAX_BUFFERSIZE // 8) + 1
        one = np.ones(1, dtype=np.int64)
        with pytest.raises(ValueError):
            blosc2.pack_array(one, clevel=-1)
        with pytest.raises(ValueError):
            blosc2.pack_array(one, clevel=10)

        # use stride trick to make an array that looks like a huge one
        # ones = np.lib.stride_tricks.as_strided(one, shape=(1, items), strides=(8, 0))[0]
        # This should always raise an error
        # FIXME: temporary disable this, as it seems that it can raise MemoryError
        #   when building wheels.  Not sure why this is happening.
        # self.assertRaises(ValueError, blosc2.pack_array, ones)

    def test_unpack_array_with_unicode_characters(self):
        import numpy as np

        input_array = np.array(["å", "ç", "ø", "π", "˚"])
        packed_array = blosc2.pack_array(input_array)
        np.testing.assert_array_equal(input_array, blosc2.unpack_array(packed_array, encoding="UTF-8"))

    def test_unpack_array_with_from_py27_exceptions(self):
        with pytest.raises(UnicodeDecodeError):
            blosc2.unpack_array(self.PY_27_INPUT)

    def test_unpack_array_with_unicode_characters_from_py27(self):
        import numpy as np

        out_array = np.array(["å", "ç", "ø", "π", "˚"])
        np.testing.assert_array_equal(out_array, blosc2.unpack_array(self.PY_27_INPUT, encoding="bytes"))

    def test_unpack_array_exceptions(self):
        with pytest.raises(TypeError):
            blosc2.unpack_array(1.0)

    @unittest.skipIf(not psutil, "psutil not available, cannot test for leaks")
    def test_no_leaks(self):
        num_elements = 10000000
        typesize = 8
        data = [float(i) for i in range(num_elements)]  # ~76MB
        Array = ctypes.c_double * num_elements
        array = Array(*data)

        def leaks(operation, repeats=3):
            gc.collect()
            used_mem_before = psutil.Process(os.getpid()).memory_info()[0]
            for _ in range(repeats):
                operation()
            gc.collect()
            used_mem_after = psutil.Process(os.getpid()).memory_info()[0]
            # We multiply by an additional factor of .01 to account for
            # storage overhead of Python classes
            return (used_mem_after - used_mem_before) >= num_elements * 8.01

        def compress():
            blosc2.compress(array, typesize, clevel=1)

        def decompress():
            cx = blosc2.compress(array, typesize, clevel=1)
            blosc2.decompress(cx)

        assert not leaks(compress), "compress leaks memory"
        assert not leaks(decompress), "decompress leaks memory"

    def test_get_blocksize(self):
        s = b"0123456789" * 1000
        blosc2.set_blocksize(2**14)
        blosc2.compress(s, typesize=1)
        d = blosc2.get_blocksize()
        assert d == 2**14

    def test_bitshuffle_not_multiple(self):
        # Check the fix for #133
        x = np.ones(27266, dtype="uint8")
        xx = x.tobytes()
        with pytest.raises(ValueError):
            blosc2.compress(xx, typesize=8, filter=blosc2.Filter.BITSHUFFLE)
        zxx = blosc2.compress(xx, typesize=1, filter=blosc2.Filter.BITSHUFFLE)
        last_xx = blosc2.decompress(zxx)[-3:]
        assert last_xx == b"\x01\x01\x01"

    def test_bitshuffle_leftovers(self):
        # Test for https://github.com/blosc2/c-blosc22/pull/100
        buffer = b" " * 641091  # a buffer that is not divisible by 8
        with pytest.raises(ValueError):
            blosc2.compress(buffer, typesize=8, filter=blosc2.Filter.BITSHUFFLE, clevel=1)
        cbuffer = blosc2.compress(buffer, typesize=1, filter=blosc2.Filter.BITSHUFFLE, clevel=1)
        dbuffer = blosc2.decompress(cbuffer)
        assert buffer == dbuffer


def run(verbosity=2):
    import blosc2.core

    blosc2.print_versions()
    suite = unittest.TestLoader().loadTestsFromTestCase(TestCodec)
    # If in the future we split this test file in several, the auto-discover
    # might be interesting

    # suite = unittest.TestLoader().discover(start_dir='.', pattern='test*.py')
    suite.addTests(unittest.TestLoader().loadTestsFromModule(blosc2.core))
    assert unittest.TextTestRunner(verbosity=verbosity).run(suite).wasSuccessful()


if __name__ == "__main__":
    run()



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import sys
from time import time

import numpy as np

import blosc2

# Dimensions, type and persistence properties for the arrays
shape = 10_000 * 10_000
chunksize = 100_000
blocksize = 10_000

dtype = np.float64

nchunks = shape // chunksize
# Set the compression and decompression parameters
cparams = blosc2.CParams(codec=blosc2.Codec.BLOSCLZ, typesize=8, blocksize=blocksize * 8)
dparams = blosc2.DParams()
contiguous = True
persistent = bool(sys.argv[1]) if len(sys.argv) > 1 else False

if persistent:
    urlpath = "bench_getitem.b2frame"
else:
    urlpath = None

storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath)
blosc2.remove_urlpath(urlpath)

# Create the empty SChunk
schunk = blosc2.SChunk(chunksize=chunksize * cparams.typesize, storage=storage, cparams=cparams, dparams=dparams)

# Append some chunks
for i in range(nchunks):
    buffer = i * np.arange(chunksize, dtype=dtype)
    nchunks_ = schunk.append_data(buffer)
    assert nchunks_ == (i + 1)

# Use get_slice for reading blocks individually
t0 = time()
for i in range(shape // blocksize):
    _ = schunk.get_slice(start=i * blocksize, stop=(i + 1) * blocksize - 1)
t1 = time()
print(f"Time for reading with get_slice: {t1 - t0:.3f}s")

blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
import math

import numpy as np
import pytest

import blosc2

argnames = "shape, chunks, blocks, slices, dtype"
argvalues = [
    ([456], [258], [73], slice(0, 1), np.int32),
    ([77, 134, 13], [31, 13, 5], [7, 8, 3], (slice(3, 7), slice(50, 100), 7), np.float64),
    ([12, 13, 14, 15, 16], [5, 5, 5, 5, 5], [2, 2, 2, 2, 2], (slice(1, 3), ..., slice(3, 6)), np.float32),
]


@pytest.mark.parametrize(argnames, argvalues)
def test_basic(shape, chunks, blocks, slices, dtype):
    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=dtype).reshape(shape)
    a = blosc2.frombuffer(bytes(nparray), nparray.shape, dtype=dtype, chunks=chunks, blocks=blocks)
    nparray_slice = nparray[slices]
    np.testing.assert_almost_equal(a[slices], nparray_slice)


@pytest.mark.parametrize(argnames, argvalues)
def test_numpy(shape, chunks, blocks, slices, dtype):
    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=dtype).reshape(shape)
    a = blosc2.asarray(nparray, chunks=chunks, blocks=blocks)
    nparray_slice = nparray[slices]
    a_slice = a[slices]

    np.testing.assert_almost_equal(a_slice, nparray_slice)


@pytest.mark.parametrize(argnames, argvalues)
def test_simple(shape, chunks, blocks, slices, dtype):
    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=dtype).reshape(shape)
    a = blosc2.asarray(nparray)
    nparray_slice = nparray[slices]
    a_slice = a[slices]

    np.testing.assert_almost_equal(a_slice, nparray_slice)


def test_shapes():
    shape = (5, 5)
    slice_ = (slice(4, 6), slice(4, 6))

    npa = np.arange(int(np.prod(shape)), dtype=np.int32).reshape(shape)
    b2a = blosc2.asarray(npa)

    # One elem slice
    assert b2a[4, 4].shape == npa[4, 4].shape
    assert b2a[4:, 4].shape == npa[4:, 4].shape
    assert b2a[4, 4:].shape == npa[4, 4:].shape
    assert b2a[4:, 4:].shape == npa[4:, 4:].shape
    assert b2a[slice_].shape == npa[slice_].shape

    # More than one elem slice
    assert b2a[3:, 4].shape == npa[3:, 4].shape
    assert b2a[3, 4:].shape == npa[3, 4:].shape
    assert b2a[3:, 4:].shape == npa[3:, 4:].shape

    # Negative values for start
    assert b2a[-1, -1].shape == npa[-1, -1].shape
    assert b2a[-1:, -2].shape == npa[-1:, -2].shape
    assert b2a[-2, -3:].shape == npa[-2, -3:].shape
    # Negative values for stop
    assert b2a[1:-1, 1].shape == npa[1:-1, 1].shape
    assert b2a[1, :-2].shape == npa[1, :-2].shape
    assert b2a[1:-2, 2:-3].shape == npa[1:-2, 2:-3].shape


def int_array(shape):
    rng = np.random.Generator(np.random.PCG64(12345))
    return rng.integers(0, shape[0], size=shape)


@pytest.mark.parametrize(
    ("shape", "chunks", "blocks", "idx"),
    [
        ((5,), (2,), (1,), int_array((2,))),
        ((15,), (4,), (2,), int_array((3,))),
        ((501,), (22,), (11,), int_array((221,))),
    ],
)
def test_1d_values(shape, chunks, blocks, idx):
    npa = np.arange(int(np.prod(shape)), dtype=np.int32).reshape(shape)
    b2a = blosc2.asarray(npa)

    np.testing.assert_equal(b2a[idx], npa[idx])
    assert b2a[idx].dtype == npa[idx].dtype
    np.testing.assert_equal(b2a[list(idx)], npa[list(idx)])
    assert b2a[list(idx)].dtype == npa[list(idx)].dtype


def bool_array(shape):
    rng = np.random.Generator(np.random.PCG64(12345))
    return rng.choice([True, False], size=shape)


@pytest.mark.parametrize(
    ("shape", "chunks", "blocks", "idx"),
    [
        ((5,), (2,), (1,), bool_array((5,))),
        ((10, 10), (5, 5), (2, 2), bool_array((10, 10))),
        ((8, 8, 8), (4, 4, 4), (2, 2, 2), bool_array((8, 8, 8))),
        ((6, 5, 4, 3), (3, 2, 2, 1), (1, 1, 1, 1), bool_array((6, 5, 4, 3))),
    ],
)
def test_bool_values(shape, chunks, blocks, idx):
    npa = np.arange(int(np.prod(shape)), dtype=np.int32).reshape(shape)
    b2a = blosc2.asarray(npa, chunks=chunks, blocks=blocks)

    assert b2a[idx].shape == npa[idx].shape
    assert b2a[idx].dtype == npa[idx].dtype
    assert b2a[idx].size == npa[idx].size
    assert b2a[idx].ndim == npa[idx].ndim


@pytest.mark.parametrize(
    ("shape", "chunks", "blocks"),
    [
        ((5,), (2,), (1,)),
        ((10, 10), (5, 5), (2, 2)),
        ((8, 8, 8), (4, 4, 4), (2, 2, 2)),
        ((6, 5, 4, 3), (3, 2, 2, 1), (1, 1, 1, 1)),
    ],
)
def test_iter(shape, chunks, blocks):
    npa = np.arange(int(np.prod(shape)), dtype=np.int32).reshape(shape)
    b2a = blosc2.asarray(npa, chunks=chunks, blocks=blocks)

    for _i, (a, b) in enumerate(zip(b2a, npa, strict=False)):
        np.testing.assert_equal(a, b)
    assert _i == shape[0] - 1


@pytest.mark.parametrize("dtype", [np.int32, np.float32, np.float64])
def test_ndarray(dtype):
    # Check that we can slice a blosc2 array with a NDArray
    shape = (10,)
    size = math.prod(shape)
    ndarray = blosc2.arange(size - 1, -1, -1, dtype=np.int64, shape=shape)
    a = blosc2.linspace(0, 10, size, shape=shape, dtype=dtype)
    a_slice = a[ndarray]
    na = np.linspace(0, 10, size, dtype=dtype).reshape(shape)
    nparray = np.arange(size - 1, -1, -1, dtype=np.int64).reshape(shape)
    na_slice = na[nparray]
    np.testing.assert_almost_equal(a_slice, na_slice)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# This shows how to evaluate expressions with NDArray instances as operands.

import numpy as np

import blosc2

shape = (10, 10, 2)

# Create a NDArray from a NumPy array
npa = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
npb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
npc = npa**2 + npb**2 + 2 * npa * npb + 1

a = blosc2.asarray(npa)
b = blosc2.asarray(npb)

# Get a LazyExpr instance
c = a**2 + b**2 + 2 * a * b + 1
# Evaluate: output is a NDArray
d = blosc2.lazyexpr("sl + c.sum() + a.std()", operands={"a": a, "c": c, "sl": a.slice((1, 1))})
print(f"Expression: {d.expression}")
print(f"Operands: {d.operands}")
assert isinstance(d, blosc2.LazyExpr)
e = d.compute()
assert isinstance(d, blosc2.LazyExpr)
# Check
assert isinstance(e, blosc2.NDArray)
sum = e[()]
print("Reduction with Blosc2:\n", sum)
npsum = npa[1, 1] + np.sum(npc) + np.std(npa)
print("Reduction with NumPy:\n", npsum)
# npsum = np.sum(npc, axis=(0,2)) + np.std(npa, axis=(0, 2))
assert np.allclose(sum, npsum)

print("NDArray expression evaluated correctly in-memory!")
